# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from swh.journal.client import JournalClient
from swh.scheduler import get_scheduler
from swh.scheduler.utils import create_task_dict


class IndexerJournalClient(JournalClient):
    """Client in charge of listing new received origins and origin_visits
       in the swh journal.

    """
    CONFIG_BASE_FILENAME = 'indexer/journal_client'

    ADDITIONAL_CONFIG = {
        'scheduler': ('dict', {
            'cls': 'remote',
            'args': {
                'url': 'http://localhost:5008/',
            }
        }),
        'origin_visit_tasks': ('List[dict]', [
            {
                'type': 'indexer_origin_metadata',
                'kwargs': {
                    'policy_update': 'update-dups',
                    'parse_ids': False,
                }
            }
        ]),
    }

    def __init__(self):
        super().__init__(extra_configuration={
            'object_types': ['origin_visit'],
        })
        self.scheduler = get_scheduler(**self.config['scheduler'])
        logging.info(
            'Starting indexer journal client with config %r',
            self.config)

    def process_objects(self, messages):
        assert set(messages) == {'origin_visit'}, set(messages)
        for origin_visit in messages['origin_visit']:
            self.process_origin_visit(origin_visit)

    def process_origin_visit(self, origin_visit):
        task_dicts = []
        logging.debug('processing origin visit %r', origin_visit)
        if origin_visit[b'status'] == b'full':
            for task_config in self.config['origin_visit_tasks']:
                logging.info(
                    'Scheduling %s for visit of origin %d',
                    task_config['type'], origin_visit[b'origin'])
                task_dicts.append(create_task_dict(
                    task_config['type'],
                    'oneshot',
                    [origin_visit[b'origin']],
                    **task_config['kwargs'],
                ))
        else:
            logging.debug('status is not "full", ignoring.')

        if task_dicts:
            self.scheduler.create_tasks(task_dicts)


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(process)d %(levelname)s %(message)s'
    )

    import click

    @click.command()
    def main():
        """Log the new received origin and origin_visits.

        """
        IndexerJournalClient().process()

    main()



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.core.api import SWHRemoteAPI

from swh.storage.exc import StorageAPIError

from .. import IndexerStorage


class RemoteStorage(SWHRemoteAPI):
    """Proxy to a remote storage API"""

    backend_class = IndexerStorage

    def __init__(self, url, timeout=None):
        super().__init__(
            api_exception=StorageAPIError, url=url, timeout=timeout)



# Copyright (C) 2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.indexer.codemeta import compact, expand
from swh.indexer.codemeta import make_absolute_uri
from swh.indexer.metadata_dictionary import MAPPINGS


def detect_metadata(files):
    """
    Detects files potentially containing metadata

    Args:
        file_entries (list): list of files

    Returns:
        dict: {mapping_filenames[name]:f['sha1']} (may be empty)
    """
    results = {}
    for (mapping_name, mapping) in MAPPINGS.items():
        matches = mapping.detect_metadata_files(files)
        if matches:
            results[mapping_name] = matches
    return results


_MINIMAL_PROPERTY_SET = {
    "developmentStatus", "version", "operatingSystem", "description",
    "keywords", "issueTracker", "name", "author", "relatedLink",
    "url", "license", "maintainer", "email", "identifier",
    "codeRepository"}

MINIMAL_METADATA_SET = {make_absolute_uri(prop)
                        for prop in _MINIMAL_PROPERTY_SET}


def extract_minimal_metadata_dict(metadata_list):
    """
    Every item in the metadata_list is a dict of translated_metadata in the
    CodeMeta vocabulary.

    We wish to extract a minimal set of terms and keep all values corresponding
    to this term without duplication.

    Args:
        metadata_list (list): list of dicts of translated_metadata

    Returns:
        dict: minimal_dict; dict with selected values of metadata
    """
    minimal_dict = {}
    for document in metadata_list:
        for metadata_item in expand(document):
            for (term, value) in metadata_item.items():
                if term in MINIMAL_METADATA_SET:
                    if term not in minimal_dict:
                        minimal_dict[term] = [value]
                    elif value not in minimal_dict[term]:
                        minimal_dict[term].append(value)
    return compact(minimal_dict)



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import subprocess

from swh.model import hashutil

from .indexer import ContentIndexer, ContentRangeIndexer, write_to_temp


def compute_license(path, log=None):
    """Determine license from file at path.

    Args:
        path: filepath to determine the license

    Returns:
        dict: A dict with the following keys:

        - licenses ([str]): associated detected licenses to path
        - path (bytes): content filepath

    """
    try:
        properties = subprocess.check_output(['nomossa', path],
                                             universal_newlines=True)
        if properties:
            res = properties.rstrip().split(' contains license(s) ')
            licenses = res[1].split(',')
        else:
            licenses = []

        return {
            'licenses': licenses,
            'path': path,
        }
    except subprocess.CalledProcessError:
        if log:
            from os import path as __path
            log.exception('Problem during license detection for sha1 %s' %
                          __path.basename(path))
        return {
            'licenses': [],
            'path': path,
        }


class MixinFossologyLicenseIndexer:
    """Mixin fossology license indexer.

    See :class:`FossologyLicenseIndexer` and
    :class:`FossologyLicenseRangeIndexer`

    """
    ADDITIONAL_CONFIG = {
        'workdir': ('str', '/tmp/swh/indexer.fossology.license'),
        'tools': ('dict', {
            'name': 'nomos',
            'version': '3.1.0rc2-31-ga2cbb8c',
            'configuration': {
                'command_line': 'nomossa <filepath>',
            },
        }),
        'write_batch_size': ('int', 1000),
    }

    CONFIG_BASE_FILENAME = 'indexer/fossology_license'

    def prepare(self):
        super().prepare()
        self.working_directory = self.config['workdir']

    def index(self, id, data):
        """Index sha1s' content and store result.

        Args:
            id (bytes): content's identifier
            raw_content (bytes): associated raw content to content id

        Returns:
            dict: A dict, representing a content_license, with keys:

            - id (bytes): content's identifier (sha1)
            - license (bytes): license in bytes
            - path (bytes): path
            - indexer_configuration_id (int): tool used to compute the output

        """
        assert isinstance(id, bytes)
        with write_to_temp(
                filename=hashutil.hash_to_hex(id),  # use the id as pathname
                data=data,
                working_directory=self.working_directory) as content_path:
            properties = compute_license(path=content_path, log=self.log)
            properties.update({
                'id': id,
                'indexer_configuration_id': self.tool['id'],
            })
        return properties

    def persist_index_computations(self, results, policy_update):
        """Persist the results in storage.

        Args:
            results ([dict]): list of content_license, dict with the
              following keys:

              - id (bytes): content's identifier (sha1)
              - license (bytes): license in bytes
              - path (bytes): path

            policy_update ([str]): either 'update-dups' or 'ignore-dups' to
              respectively update duplicates or ignore them

        """
        self.idx_storage.content_fossology_license_add(
            results, conflict_update=(policy_update == 'update-dups'))


class FossologyLicenseIndexer(
        MixinFossologyLicenseIndexer, ContentIndexer):
    """Indexer in charge of:

    - filtering out content already indexed
    - reading content from objstorage per the content's id (sha1)
    - computing {license, encoding} from that content
    - store result in storage

    """
    def filter(self, ids):
        """Filter out known sha1s and return only missing ones.

        """
        yield from self.idx_storage.content_fossology_license_missing((
            {
                'id': sha1,
                'indexer_configuration_id': self.tool['id'],
            } for sha1 in ids
        ))


class FossologyLicenseRangeIndexer(
        MixinFossologyLicenseIndexer, ContentRangeIndexer):
    """FossologyLicense Range Indexer working on range of content identifiers.

    - filters out the non textual content
    - (optionally) filters out content already indexed (cf
      :meth:`.indexed_contents_in_range`)
    - reads content from objstorage per the content's id (sha1)
    - computes {mimetype, encoding} from that content
    - stores result in storage

    """
    def indexed_contents_in_range(self, start, end):
        """Retrieve indexed content id within range [start, end].

        Args:
            start (bytes): Starting bound from range identifier
            end (bytes): End range identifier

        Returns:
            dict: a dict with keys:

            - **ids** [bytes]: iterable of content ids within the range.
            - **next** (Optional[bytes]): The next range of sha1 starts at
              this sha1 if any

        """
        return self.idx_storage.content_fossology_license_get_range(
                start, end, self.tool['id'])



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import bisect
from collections import defaultdict, Counter
import itertools
import json
import operator
import math
import re

from . import MAPPING_NAMES

SHA1_DIGEST_SIZE = 160


def _transform_tool(tool):
    return {
        'id': tool['id'],
        'name': tool['tool_name'],
        'version': tool['tool_version'],
        'configuration': tool['tool_configuration'],
    }


class SubStorage:
    """Implements common missing/get/add logic for each indexer type."""
    def __init__(self, tools):
        self._tools = tools
        self._sorted_ids = []
        self._data = {}  # map (id_, tool_id) -> metadata_dict
        self._tools_per_id = defaultdict(set)  # map id_ -> Set[tool_id]

    def missing(self, ids):
        """List data missing from storage.

        Args:
            data (iterable): dictionaries with keys:

                - **id** (bytes): sha1 identifier
                - **indexer_configuration_id** (int): tool used to compute
                  the results

        Yields:
            missing sha1s

        """
        for id_ in ids:
            tool_id = id_['indexer_configuration_id']
            id_ = id_['id']
            if tool_id not in self._tools_per_id.get(id_, set()):
                yield id_

    def get(self, ids):
        """Retrieve data per id.

        Args:
            ids (iterable): sha1 checksums

        Yields:
            dict: dictionaries with the following keys:

              - **id** (bytes)
              - **tool** (dict): tool used to compute metadata
              - arbitrary data (as provided to `add`)

        """
        for id_ in ids:
            for tool_id in self._tools_per_id.get(id_, set()):
                key = (id_, tool_id)
                yield {
                    'id': id_,
                    'tool': _transform_tool(self._tools[tool_id]),
                    **self._data[key],
                }

    def get_all(self):
        yield from self.get(list(self._tools_per_id))

    def get_range(self, start, end, indexer_configuration_id, limit):
        """Retrieve data within range [start, end] bound by limit.

        Args:
            **start** (bytes): Starting identifier range (expected smaller
                           than end)
            **end** (bytes): Ending identifier range (expected larger
                             than start)
            **indexer_configuration_id** (int): The tool used to index data
            **limit** (int): Limit result

        Raises:
            ValueError for limit to None

        Returns:
            a dict with keys:
            - **ids** [bytes]: iterable of content ids within the range.
            - **next** (Optional[bytes]): The next range of sha1 starts at
                                          this sha1 if any

        """
        if limit is None:
            raise ValueError('Development error: limit should not be None')
        from_index = bisect.bisect_left(self._sorted_ids, start)
        to_index = bisect.bisect_right(self._sorted_ids, end, lo=from_index)
        if to_index - from_index >= limit:
            return {
                'ids': self._sorted_ids[from_index:from_index+limit],
                'next': self._sorted_ids[from_index+limit],
            }
        else:
            return {
                'ids': self._sorted_ids[from_index:to_index],
                'next': None,
                }

    def add(self, data, conflict_update):
        """Add data not present in storage.

        Args:
            data (iterable): dictionaries with keys:

              - **id**: sha1
              - **indexer_configuration_id**: tool used to compute the
                results
              - arbitrary data

            conflict_update (bool): Flag to determine if we want to overwrite
              (true) or skip duplicates (false)

        """
        data = list(data)
        if len({x['id'] for x in data}) < len(data):
            # For "exception-compatibility" with the pgsql backend
            raise ValueError('The same id is present more than once.')
        for item in data:
            item = item.copy()
            tool_id = item.pop('indexer_configuration_id')
            id_ = item.pop('id')
            data = item
            if not conflict_update and \
                    tool_id in self._tools_per_id.get(id_, set()):
                # Duplicate, should not be updated
                continue
            key = (id_, tool_id)
            self._data[key] = data
            self._tools_per_id[id_].add(tool_id)
            if id_ not in self._sorted_ids:
                bisect.insort(self._sorted_ids, id_)

    def add_merge(self, new_data, conflict_update, merged_key):
        for new_item in new_data:
            id_ = new_item['id']
            tool_id = new_item['indexer_configuration_id']
            if conflict_update:
                all_subitems = []
            else:
                existing = list(self.get([id_]))
                all_subitems = [
                    old_subitem
                    for existing_item in existing
                    if existing_item['tool']['id'] == tool_id
                    for old_subitem in existing_item[merged_key]
                ]
            for new_subitem in new_item[merged_key]:
                if new_subitem not in all_subitems:
                    all_subitems.append(new_subitem)
            self.add([
                {
                    'id': id_,
                    'indexer_configuration_id': tool_id,
                    merged_key: all_subitems,
                }
            ], conflict_update=True)
            if id_ not in self._sorted_ids:
                bisect.insort(self._sorted_ids, id_)


class IndexerStorage:
    """In-memory SWH indexer storage."""

    def __init__(self):
        self._tools = {}
        self._mimetypes = SubStorage(self._tools)
        self._languages = SubStorage(self._tools)
        self._content_ctags = SubStorage(self._tools)
        self._licenses = SubStorage(self._tools)
        self._content_metadata = SubStorage(self._tools)
        self._revision_metadata = SubStorage(self._tools)
        self._origin_intrinsic_metadata = SubStorage(self._tools)

    def content_mimetype_missing(self, mimetypes):
        """Generate mimetypes missing from storage.

        Args:
            mimetypes (iterable): iterable of dict with keys:

              - **id** (bytes): sha1 identifier
              - **indexer_configuration_id** (int): tool used to compute the
                results

        Yields:
            tuple (id, indexer_configuration_id): missing id

        """
        yield from self._mimetypes.missing(mimetypes)

    def content_mimetype_get_range(
            self, start, end, indexer_configuration_id, limit=1000):
        """Retrieve mimetypes within range [start, end] bound by limit.

        Args:
            **start** (bytes): Starting identifier range (expected smaller
                           than end)
            **end** (bytes): Ending identifier range (expected larger
                             than start)
            **indexer_configuration_id** (int): The tool used to index data
            **limit** (int): Limit result (default to 1000)

        Raises:
            ValueError for limit to None

        Returns:
            a dict with keys:
            - **ids** [bytes]: iterable of content ids within the range.
            - **next** (Optional[bytes]): The next range of sha1 starts at
                                          this sha1 if any

        """
        return self._mimetypes.get_range(
            start, end, indexer_configuration_id, limit)

    def content_mimetype_add(self, mimetypes, conflict_update=False):
        """Add mimetypes not present in storage.

        Args:
            mimetypes (iterable): dictionaries with keys:

              - **id** (bytes): sha1 identifier
              - **mimetype** (bytes): raw content's mimetype
              - **encoding** (bytes): raw content's encoding
              - **indexer_configuration_id** (int): tool's id used to
                compute the results
              - **conflict_update** (bool): Flag to determine if we want to
                overwrite (``True``) or skip duplicates (``False``, the
                default)

        """
        if not all(isinstance(x['id'], bytes) for x in mimetypes):
            raise TypeError('identifiers must be bytes.')
        self._mimetypes.add(mimetypes, conflict_update)

    def content_mimetype_get(self, ids, db=None, cur=None):
        """Retrieve full content mimetype per ids.

        Args:
            ids (iterable): sha1 identifier

        Yields:
            mimetypes (iterable): dictionaries with keys:

                - **id** (bytes): sha1 identifier
                - **mimetype** (bytes): raw content's mimetype
                - **encoding** (bytes): raw content's encoding
                - **tool** (dict): Tool used to compute the language

        """
        yield from self._mimetypes.get(ids)

    def content_language_missing(self, languages):
        """List languages missing from storage.

        Args:
            languages (iterable): dictionaries with keys:

                - **id** (bytes): sha1 identifier
                - **indexer_configuration_id** (int): tool used to compute
                  the results

        Yields:
            an iterable of missing id for the tuple (id,
            indexer_configuration_id)

        """
        yield from self._languages.missing(languages)

    def content_language_get(self, ids):
        """Retrieve full content language per ids.

        Args:
            ids (iterable): sha1 identifier

        Yields:
            languages (iterable): dictionaries with keys:

                - **id** (bytes): sha1 identifier
                - **lang** (bytes): raw content's language
                - **tool** (dict): Tool used to compute the language

        """
        yield from self._languages.get(ids)

    def content_language_add(self, languages, conflict_update=False):
        """Add languages not present in storage.

        Args:
            languages (iterable): dictionaries with keys:

                - **id** (bytes): sha1
                - **lang** (bytes): language detected

            conflict_update (bool): Flag to determine if we want to
                overwrite (true) or skip duplicates (false, the
                default)

        """
        if not all(isinstance(x['id'], bytes) for x in languages):
            raise TypeError('identifiers must be bytes.')
        self._languages.add(languages, conflict_update)

    def content_ctags_missing(self, ctags):
        """List ctags missing from storage.

        Args:
            ctags (iterable): dicts with keys:

                - **id** (bytes): sha1 identifier
                - **indexer_configuration_id** (int): tool used to compute
                  the results

        Yields:
            an iterable of missing id for the tuple (id,
            indexer_configuration_id)

        """
        yield from self._content_ctags.missing(ctags)

    def content_ctags_get(self, ids):
        """Retrieve ctags per id.

        Args:
            ids (iterable): sha1 checksums

        Yields:
            Dictionaries with keys:

                - **id** (bytes): content's identifier
                - **name** (str): symbol's name
                - **kind** (str): symbol's kind
                - **lang** (str): language for that content
                - **tool** (dict): tool used to compute the ctags' info


        """
        for item in self._content_ctags.get(ids):
            for item_ctags_item in item['ctags']:
                yield {
                    'id': item['id'],
                    'tool': item['tool'],
                    **item_ctags_item
                }

    def content_ctags_add(self, ctags, conflict_update=False):
        """Add ctags not present in storage

        Args:
            ctags (iterable): dictionaries with keys:

              - **id** (bytes): sha1
              - **ctags** ([list): List of dictionary with keys: name, kind,
                  line, lang
              - **indexer_configuration_id**: tool used to compute the
                results

        """
        if not all(isinstance(x['id'], bytes) for x in ctags):
            raise TypeError('identifiers must be bytes.')
        self._content_ctags.add_merge(ctags, conflict_update, 'ctags')

    def content_ctags_search(self, expression,
                             limit=10, last_sha1=None, db=None, cur=None):
        """Search through content's raw ctags symbols.

        Args:
            expression (str): Expression to search for
            limit (int): Number of rows to return (default to 10).
            last_sha1 (str): Offset from which retrieving data (default to '').

        Yields:
            rows of ctags including id, name, lang, kind, line, etc...

        """
        nb_matches = 0
        for ((id_, tool_id), item) in \
                sorted(self._content_ctags._data.items()):
            if id_ <= (last_sha1 or bytes(0 for _ in range(SHA1_DIGEST_SIZE))):
                continue
            for ctags_item in item['ctags']:
                if ctags_item['name'] != expression:
                    continue
                nb_matches += 1
                yield {
                    'id': id_,
                    'tool': _transform_tool(self._tools[tool_id]),
                    **ctags_item
                }
                if nb_matches >= limit:
                    return

    def content_fossology_license_get(self, ids):
        """Retrieve licenses per id.

        Args:
            ids (iterable): sha1 checksums

        Yields:
            `{id: facts}` where `facts` is a dict with the following keys:

                - **licenses** ([str]): associated licenses for that content
                - **tool** (dict): Tool used to compute the license

        """
        # Rewrites the output of SubStorage.get from the old format to
        # the new one. SubStorage.get should be updated once all other
        # *_get methods use the new format.
        # See: https://forge.softwareheritage.org/T1433
        res = {}
        for d in self._licenses.get(ids):
            res.setdefault(d.pop('id'), []).append(d)
        for (id_, facts) in res.items():
            yield {id_: facts}

    def content_fossology_license_add(self, licenses, conflict_update=False):
        """Add licenses not present in storage.

        Args:
            licenses (iterable): dictionaries with keys:

                - **id**: sha1
                - **licenses** ([bytes]): List of licenses associated to sha1
                - **tool** (str): nomossa

            conflict_update: Flag to determine if we want to overwrite (true)
                or skip duplicates (false, the default)

        Returns:
            list: content_license entries which failed due to unknown licenses

        """
        if not all(isinstance(x['id'], bytes) for x in licenses):
            raise TypeError('identifiers must be bytes.')
        self._licenses.add_merge(licenses, conflict_update, 'licenses')

    def content_fossology_license_get_range(
            self, start, end, indexer_configuration_id, limit=1000):
        """Retrieve licenses within range [start, end] bound by limit.

        Args:
            **start** (bytes): Starting identifier range (expected smaller
                           than end)
            **end** (bytes): Ending identifier range (expected larger
                             than start)
            **indexer_configuration_id** (int): The tool used to index data
            **limit** (int): Limit result (default to 1000)

        Raises:
            ValueError for limit to None

        Returns:
            a dict with keys:
            - **ids** [bytes]: iterable of content ids within the range.
            - **next** (Optional[bytes]): The next range of sha1 starts at
                                          this sha1 if any

        """
        return self._licenses.get_range(
            start, end, indexer_configuration_id, limit)

    def content_metadata_missing(self, metadata):
        """List metadata missing from storage.

        Args:
            metadata (iterable): dictionaries with keys:

              - **id** (bytes): sha1 identifier
              - **indexer_configuration_id** (int): tool used to compute
                the results

        Yields:
            missing sha1s

        """
        yield from self._content_metadata.missing(metadata)

    def content_metadata_get(self, ids):
        """Retrieve metadata per id.

        Args:
            ids (iterable): sha1 checksums

        Yields:
            dictionaries with the following keys:

              - **id** (bytes)
              - **translated_metadata** (str): associated metadata
              - **tool** (dict): tool used to compute metadata

        """
        yield from self._content_metadata.get(ids)

    def content_metadata_add(self, metadata, conflict_update=False):
        """Add metadata not present in storage.

        Args:
            metadata (iterable): dictionaries with keys:

              - **id**: sha1
              - **translated_metadata**: arbitrary dict
              - **indexer_configuration_id**: tool used to compute the
                results

            conflict_update: Flag to determine if we want to overwrite (true)
                or skip duplicates (false, the default)

        """
        if not all(isinstance(x['id'], bytes) for x in metadata):
            raise TypeError('identifiers must be bytes.')
        self._content_metadata.add(metadata, conflict_update)

    def revision_metadata_missing(self, metadata):
        """List metadata missing from storage.

        Args:
            metadata (iterable): dictionaries with keys:

              - **id** (bytes): sha1_git revision identifier
              - **indexer_configuration_id** (int): tool used to compute
                the results

        Yields:
            missing ids

        """
        yield from self._revision_metadata.missing(metadata)

    def revision_metadata_get(self, ids):
        """Retrieve revision metadata per id.

        Args:
            ids (iterable): sha1 checksums

        Yields:
            dictionaries with the following keys:

            - **id** (bytes)
            - **translated_metadata** (str): associated metadata
            - **tool** (dict): tool used to compute metadata
            - **mappings** (List[str]): list of mappings used to translate
              these metadata

        """
        yield from self._revision_metadata.get(ids)

    def revision_metadata_add(self, metadata, conflict_update=False):
        """Add metadata not present in storage.

        Args:
            metadata (iterable): dictionaries with keys:

              - **id**: sha1_git of revision
              - **translated_metadata**: arbitrary dict
              - **indexer_configuration_id**: tool used to compute metadata
              - **mappings** (List[str]): list of mappings used to translate
                these metadata

            conflict_update: Flag to determine if we want to overwrite (true)
              or skip duplicates (false, the default)

        """
        if not all(isinstance(x['id'], bytes) for x in metadata):
            raise TypeError('identifiers must be bytes.')
        self._revision_metadata.add(metadata, conflict_update)

    def origin_intrinsic_metadata_get(self, ids):
        """Retrieve origin metadata per id.

        Args:
            ids (iterable): origin identifiers

        Yields:
            list: dictionaries with the following keys:

                - **origin_id** (int)
                - **translated_metadata** (str): associated metadata
                - **tool** (dict): tool used to compute metadata
                - **mappings** (List[str]): list of mappings used to translate
                  these metadata

        """
        for item in self._origin_intrinsic_metadata.get(ids):
            item['origin_id'] = item.pop('id')
            yield item

    def origin_intrinsic_metadata_add(self, metadata,
                                      conflict_update=False):
        """Add origin metadata not present in storage.

        Args:
            metadata (iterable): dictionaries with keys:

                - **origin_id**: origin identifier
                - **from_revision**: sha1 id of the revision used to generate
                  these metadata.
                - **metadata**: arbitrary dict
                - **indexer_configuration_id**: tool used to compute metadata
                - **mappings** (List[str]): list of mappings used to translate
                  these metadata

            conflict_update: Flag to determine if we want to overwrite (true)
              or skip duplicates (false, the default)

        """

        items = []
        for item in metadata:
            item = item.copy()
            item['id'] = item.pop('origin_id')
            items.append(item)
        self._origin_intrinsic_metadata.add(items, conflict_update)

    def origin_intrinsic_metadata_search_fulltext(
            self, conjunction, limit=100):
        """Returns the list of origins whose metadata contain all the terms.

        Args:
            conjunction (List[str]): List of terms to be searched for.
            limit (int): The maximum number of results to return

        Yields:
            list: dictionaries with the following keys:

                - **id** (int)
                - **metadata** (str): associated metadata
                - **tool** (dict): tool used to compute metadata
                - **mappings** (List[str]): list of mappings used to translate
                  these metadata

        """
        # A very crude fulltext search implementation, but that's enough
        # to work on English metadata
        tokens_re = re.compile('[a-zA-Z0-9]+')
        search_tokens = list(itertools.chain(
            *map(tokens_re.findall, conjunction)))

        def rank(data):
            # Tokenize the metadata
            text = json.dumps(data['metadata'])
            text_tokens = tokens_re.findall(text)
            text_token_occurences = Counter(text_tokens)

            # Count the number of occurrences of search tokens in the text
            score = 0
            for search_token in search_tokens:
                if text_token_occurences[search_token] == 0:
                    # Search token is not in the text.
                    return 0
                score += text_token_occurences[search_token]

            # Normalize according to the text's length
            return score / math.log(len(text_tokens))

        results = [(rank(data), data)
                   for data in self._origin_intrinsic_metadata.get_all()]
        results = [(rank_, data) for (rank_, data) in results if rank_ > 0]
        results.sort(key=operator.itemgetter(0),  # Don't try to order 'data'
                     reverse=True)
        for (rank_, result) in results[:limit]:
            result = result.copy()
            result['origin_id'] = result.pop('id')
            yield result

    def origin_intrinsic_metadata_search_by_producer(
            self, start=0, end=None, limit=100, ids_only=False,
            mappings=None, tool_ids=None,
            db=None, cur=None):
        """Returns the list of origins whose metadata contain all the terms.

        Args:
            start (int): The minimum origin id to return
            end (int): The maximum origin id to return
            limit (int): The maximum number of results to return
            ids_only (bool): Determines whether only origin ids are returned
                or the content as well
            mappings (List[str]): Returns origins whose intrinsic metadata
                were generated using at least one of these mappings.

        Yields:
            list: list of origin ids (int) if `ids_only=True`, else
                dictionaries with the following keys:

                - **id** (int)
                - **metadata** (str): associated metadata
                - **tool** (dict): tool used to compute metadata
                - **mappings** (List[str]): list of mappings used to translate
                  these metadata

        """
        nb_results = 0
        if mappings is not None:
            mappings = frozenset(mappings)
        if tool_ids is not None:
            tool_ids = frozenset(tool_ids)
        for entry in self._origin_intrinsic_metadata.get_all():
            if entry['id'] < start or (end and entry['id'] > end):
                continue
            if nb_results >= limit:
                return
            if mappings is not None and mappings.isdisjoint(entry['mappings']):
                continue
            if tool_ids is not None and entry['tool']['id'] not in tool_ids:
                continue
            if ids_only:
                yield entry['id']
            else:
                entry = entry.copy()
                entry['origin_id'] = entry.pop('id')
                yield entry
            nb_results += 1

    def origin_intrinsic_metadata_stats(self):
        """Returns statistics on stored intrinsic metadata.

        Returns:
            dict: dictionary with keys:

                - total (int): total number of origins that were indexed
                  (possibly yielding an empty metadata dictionary)
                - non_empty (int): total number of origins that we extracted
                  a non-empty metadata dictionary from
                - per_mapping (dict): a dictionary with mapping names as
                  keys and number of origins whose indexing used this
                  mapping. Note that indexing a given origin may use
                  0, 1, or many mappings.
        """
        mapping_count = {m: 0 for m in MAPPING_NAMES}
        total = non_empty = 0
        for data in self._origin_intrinsic_metadata.get_all():
            total += 1
            if set(data['metadata']) - {'@context'}:
                non_empty += 1
            for mapping in data['mappings']:
                mapping_count[mapping] += 1
        return {
            'per_mapping': mapping_count,
            'total': total,
            'non_empty': non_empty
        }

    def indexer_configuration_add(self, tools):
        """Add new tools to the storage.

        Args:
            tools ([dict]): List of dictionary representing tool to
              insert in the db. Dictionary with the following keys:

              - **tool_name** (str): tool's name
              - **tool_version** (str): tool's version
              - **tool_configuration** (dict): tool's configuration
                (free form dict)

        Returns:
            list: List of dict inserted in the db (holding the id key as
            well). The order of the list is not guaranteed to match
            the order of the initial list.

        """
        inserted = []
        for tool in tools:
            tool = tool.copy()
            id_ = self._tool_key(tool)
            tool['id'] = id_
            self._tools[id_] = tool
            inserted.append(tool)
        return inserted

    def indexer_configuration_get(self, tool):
        """Retrieve tool information.

        Args:
            tool (dict): Dictionary representing a tool with the
              following keys:

              - **tool_name** (str): tool's name
              - **tool_version** (str): tool's version
              - **tool_configuration** (dict): tool's configuration
                (free form dict)

        Returns:
            The same dictionary with an `id` key, None otherwise.

        """
        return self._tools.get(self._tool_key(tool))

    def _tool_key(self, tool):
        return hash((tool['tool_name'], tool['tool_version'],
                     json.dumps(tool['tool_configuration'], sort_keys=True)))



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import magic

from swh.model import hashutil

from .indexer import ContentIndexer, ContentRangeIndexer


def compute_mimetype_encoding(raw_content):
    """Determine mimetype and encoding from the raw content.

    Args:
        raw_content (bytes): content's raw data

    Returns:
        dict: mimetype and encoding key and corresponding values
        (as bytes).

    """
    r = magic.detect_from_content(raw_content)
    return {
        'mimetype': r.mime_type,
        'encoding': r.encoding,
    }


class MixinMimetypeIndexer:
    """Mixin mimetype indexer.

    See :class:`MimetypeIndexer` and :class:`MimetypeRangeIndexer`

    """
    ADDITIONAL_CONFIG = {
        'tools': ('dict', {
            'name': 'file',
            'version': '1:5.30-1+deb9u1',
            'configuration': {
                "type": "library",
                "debian-package": "python3-magic"
            },
        }),
        'write_batch_size': ('int', 1000),
    }

    CONFIG_BASE_FILENAME = 'indexer/mimetype'

    def index(self, id, data):
        """Index sha1s' content and store result.

        Args:
            id (bytes): content's identifier
            data (bytes): raw content in bytes

        Returns:
            dict: content's mimetype; dict keys being

            - **id** (bytes): content's identifier (sha1)
            - **mimetype** (bytes): mimetype in bytes
            - **encoding** (bytes): encoding in bytes

        """
        try:
            properties = compute_mimetype_encoding(data)
            properties.update({
                'id': id,
                'indexer_configuration_id': self.tool['id'],
                })
        except TypeError:
            self.log.error('Detecting mimetype error for id %s' % (
                hashutil.hash_to_hex(id), ))
            return None

        return properties

    def persist_index_computations(self, results, policy_update):
        """Persist the results in storage.

        Args:
            results ([dict]): list of content's mimetype dicts
              (see :meth:`.index`)

            policy_update ([str]): either 'update-dups' or 'ignore-dups' to
               respectively update duplicates or ignore them

        """
        self.idx_storage.content_mimetype_add(
            results, conflict_update=(policy_update == 'update-dups'))


class MimetypeIndexer(MixinMimetypeIndexer, ContentIndexer):
    """Mimetype Indexer working on list of content identifiers.

    It:

    - (optionally) filters out content already indexed (cf.
      :meth:`.filter`)
    - reads content from objstorage per the content's id (sha1)
    - computes {mimetype, encoding} from that content
    - stores result in storage

    """
    def filter(self, ids):
        """Filter out known sha1s and return only missing ones.

        """
        yield from self.idx_storage.content_mimetype_missing((
            {
                'id': sha1,
                'indexer_configuration_id': self.tool['id'],
            } for sha1 in ids
        ))


class MimetypeRangeIndexer(MixinMimetypeIndexer, ContentRangeIndexer):
    """Mimetype Range Indexer working on range of content identifiers.

    It:

    - (optionally) filters out content already indexed (cf
      :meth:`.indexed_contents_in_range`)
    - reads content from objstorage per the content's id (sha1)
    - computes {mimetype, encoding} from that content
    - stores result in storage

    """
    def indexed_contents_in_range(self, start, end):
        """Retrieve indexed content id within range [start, end].

        Args:
            start (bytes): Starting bound from range identifier
            end (bytes): End range identifier

        Returns:
            dict: a dict with keys:

            - **ids** [bytes]: iterable of content ids within the range.
            - **next** (Optional[bytes]): The next range of sha1 starts at
              this sha1 if any

        """
        return self.idx_storage.content_mimetype_get_range(
            start, end, self.tool['id'])



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from unittest.mock import patch

import pytest

from swh.indexer import fossology_license
from swh.indexer.fossology_license import (
    FossologyLicenseIndexer, FossologyLicenseRangeIndexer,
    compute_license
)

from swh.indexer.tests.utils import (
    SHA1_TO_LICENSES, CommonContentIndexerTest, CommonContentIndexerRangeTest,
    BASE_TEST_CONFIG, fill_storage, fill_obj_storage, filter_dict,
)


class BasicTest(unittest.TestCase):
    @patch('swh.indexer.fossology_license.subprocess')
    def test_compute_license(self, mock_subprocess):
        """Computing licenses from a raw content should return results

        """
        for path, intermediary_result, output in [
                (b'some/path', None,
                 []),
                (b'some/path/2', [],
                 []),
                (b'other/path', ' contains license(s) GPL,AGPL',
                 ['GPL', 'AGPL'])]:
            mock_subprocess.check_output.return_value = intermediary_result

            actual_result = compute_license(path, log=None)

            self.assertEqual(actual_result, {
                'licenses': output,
                'path': path,
            })


def mock_compute_license(path, log=None):
    """path is the content identifier

    """
    if isinstance(id, bytes):
        path = path.decode('utf-8')
    # path is something like /tmp/tmpXXX/<sha1> so we keep only the sha1 part
    path = path.split('/')[-1]
    return {
        'licenses': SHA1_TO_LICENSES.get(path)
    }


CONFIG = {
    **BASE_TEST_CONFIG,
    'workdir': '/tmp',
    'tools': {
        'name': 'nomos',
        'version': '3.1.0rc2-31-ga2cbb8c',
        'configuration': {
            'command_line': 'nomossa <filepath>',
        },
    },
}

RANGE_CONFIG = dict(list(CONFIG.items()) + [('write_batch_size', 100)])


class TestFossologyLicenseIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Language indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_fossology_license_get(ids)

    def setUp(self):
        super().setUp()
        # replace actual license computation with a mock
        self.orig_compute_license = fossology_license.compute_license
        fossology_license.compute_license = mock_compute_license

        self.indexer = FossologyLicenseIndexer(CONFIG)
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0 = '01c9379dfc33803963d07c1ccc748d3fe4c96bb5'
        self.id1 = '688a5ef812c53907562fe379d4b3851e69c7cb15'
        self.id2 = 'da39a3ee5e6b4b0d3255bfef95601890afd80709'  # empty content

        tool = {k.replace('tool_', ''): v
                for (k, v) in self.indexer.tool.items()}
        # then
        self.expected_results = {
            self.id0: {
                'tool': tool,
                'licenses': SHA1_TO_LICENSES[self.id0],
            },
            self.id1: {
                'tool': tool,
                'licenses': SHA1_TO_LICENSES[self.id1],
            },
            self.id2: {
                'tool': tool,
                'licenses': SHA1_TO_LICENSES[self.id2],
            }
        }

    def tearDown(self):
        super().tearDown()
        fossology_license.compute_license = self.orig_compute_license


class TestFossologyLicenseRangeIndexer(
        CommonContentIndexerRangeTest, unittest.TestCase):
    """Range Fossology License Indexer tests.

    - new data within range are indexed
    - no data outside a range are indexed
    - with filtering existing indexed data prior to compute new index
    - without filtering existing indexed data prior to compute new index

    """
    def setUp(self):
        super().setUp()

        # replace actual license computation with a mock
        self.orig_compute_license = fossology_license.compute_license
        fossology_license.compute_license = mock_compute_license

        self.indexer = FossologyLicenseRangeIndexer(config=RANGE_CONFIG)
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0 = '01c9379dfc33803963d07c1ccc748d3fe4c96bb5'
        self.id1 = '02fb2c89e14f7fab46701478c83779c7beb7b069'
        self.id2 = '103bc087db1d26afc3a0283f38663d081e9b01e6'
        tool_id = self.indexer.tool['id']
        self.expected_results = {
            self.id0: {
                'id': self.id0,
                'indexer_configuration_id': tool_id,
                'licenses': SHA1_TO_LICENSES[self.id0]
            },
            self.id1: {
                'id': self.id1,
                'indexer_configuration_id': tool_id,
                'licenses': SHA1_TO_LICENSES[self.id1]
            },
            self.id2: {
                'id': self.id2,
                'indexer_configuration_id': tool_id,
                'licenses': SHA1_TO_LICENSES[self.id2]
            }
        }

    def tearDown(self):
        super().tearDown()
        fossology_license.compute_license = self.orig_compute_license


def test_fossology_w_no_tool():
    with pytest.raises(ValueError):
        FossologyLicenseIndexer(config=filter_dict(CONFIG, 'tools'))


def test_fossology_range_w_no_tool():
    with pytest.raises(ValueError):
        FossologyLicenseRangeIndexer(config=filter_dict(RANGE_CONFIG, 'tools'))



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.indexer.origin_head import OriginHeadIndexer
from swh.indexer.tests.utils import (
    BASE_TEST_CONFIG, fill_storage
)

ORIGIN_HEAD_CONFIG = {
    **BASE_TEST_CONFIG,
    'tools': {
        'name': 'origin-metadata',
        'version': '0.0.1',
        'configuration': {},
    },
    'tasks': {
        'revision_metadata': None,
        'origin_intrinsic_metadata': None,
    }
}


class OriginHeadTestIndexer(OriginHeadIndexer):
    """Specific indexer whose configuration is enough to satisfy the
       indexing tests.
    """
    def parse_config_file(self, *args, **kwargs):
        return ORIGIN_HEAD_CONFIG

    def persist_index_computations(self, results, policy_update):
        self.results = results


class OriginHead(unittest.TestCase):
    def setUp(self):
        self.indexer = OriginHeadTestIndexer()
        fill_storage(self.indexer.storage)

    def _get_origin_id(self, type_, url):
        origin = self.indexer.storage.origin_get({
            'type': type_, 'url': url})
        return origin['id']

    def test_git(self):
        self.indexer.run(
                ['git+https://github.com/SoftwareHeritage/swh-storage'])
        origin_id = self._get_origin_id(
            'git', 'https://github.com/SoftwareHeritage/swh-storage')
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'8K\x12\x00d\x03\xcc\xe4]bS\xe3\x8f{'
                           b'\xd7}\xac\xefrm',
            'origin_id': origin_id}])

    def test_ftp(self):
        self.indexer.run(
                ['ftp+rsync://ftp.gnu.org/gnu/3dldf'])
        origin_id = self._get_origin_id(
                'ftp', 'rsync://ftp.gnu.org/gnu/3dldf')
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\x8e\xa9\x8e/\xea}\x9feF\xf4\x9f\xfd\xee'
                           b'\xcc\x1a\xb4`\x8c\x8by',
            'origin_id': origin_id}])

    def test_deposit(self):
        self.indexer.run(
                ['deposit+https://forge.softwareheritage.org/source/'
                 'jesuisgpl/'])
        origin_id = self._get_origin_id(
            'deposit', 'https://forge.softwareheritage.org/source/jesuisgpl/')
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\xe7n\xa4\x9c\x9f\xfb\xb7\xf76\x11\x08{'
                           b'\xa6\xe9\x99\xb1\x9e]q\xeb',
            'origin_id': origin_id}])

    def test_pypi(self):
        self.indexer.run(
                ['pypi+https://pypi.org/project/limnoria/'])
        origin_id = self._get_origin_id(
            'pypi', 'https://pypi.org/project/limnoria/')
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8k'
                           b'A\x10\x9d\xc5\xfa2\xf8t',
            'origin_id': origin_id}])

    def test_svn(self):
        self.indexer.run(
                ['svn+http://0-512-md.googlecode.com/svn/'])
        origin_id = self._get_origin_id(
            'svn', 'http://0-512-md.googlecode.com/svn/')
        self.assertEqual(self.indexer.results, [{
            'revision_id': b'\xe4?r\xe1,\x88\xab\xec\xe7\x9a\x87\xb8'
                           b'\xc9\xad#.\x1bw=\x18',
            'origin_id': origin_id}])



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.indexer.storage import converters


class TestConverters(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_ctags_to_db(self):
        input_ctag = {
            'id': b'some-id',
            'indexer_configuration_id': 100,
            'ctags': [
                {
                    'name': 'some-name',
                    'kind': 'some-kind',
                    'line': 10,
                    'lang': 'Yaml',
                }, {
                    'name': 'main',
                    'kind': 'function',
                    'line': 12,
                    'lang': 'Yaml',
                },
            ]
        }

        expected_ctags = [
            {
                'id': b'some-id',
                'name': 'some-name',
                'kind': 'some-kind',
                'line': 10,
                'lang': 'Yaml',
                'indexer_configuration_id': 100,
            }, {
                'id': b'some-id',
                'name': 'main',
                'kind': 'function',
                'line': 12,
                'lang': 'Yaml',
                'indexer_configuration_id': 100,
            }]

        # when
        actual_ctags = list(converters.ctags_to_db(input_ctag))

        # then
        self.assertEqual(actual_ctags, expected_ctags)

    def test_db_to_ctags(self):
        input_ctags = {
            'id': b'some-id',
            'name': 'some-name',
            'kind': 'some-kind',
            'line': 10,
            'lang': 'Yaml',
            'tool_id': 200,
            'tool_name': 'some-toolname',
            'tool_version': 'some-toolversion',
            'tool_configuration': {}
        }
        expected_ctags = {
            'id': b'some-id',
            'name': 'some-name',
            'kind': 'some-kind',
            'line': 10,
            'lang': 'Yaml',
            'tool': {
                'id': 200,
                'name': 'some-toolname',
                'version': 'some-toolversion',
                'configuration': {},
            }
        }

        # when
        actual_ctags = converters.db_to_ctags(input_ctags)

        # then
        self.assertEqual(actual_ctags, expected_ctags)

    def test_db_to_mimetype(self):
        input_mimetype = {
            'id': b'some-id',
            'tool_id': 10,
            'tool_name': 'some-toolname',
            'tool_version': 'some-toolversion',
            'tool_configuration': {},
            'encoding': b'ascii',
            'mimetype': b'text/plain',
        }

        expected_mimetype = {
            'id': b'some-id',
            'encoding': b'ascii',
            'mimetype': b'text/plain',
            'tool': {
                'id': 10,
                'name': 'some-toolname',
                'version': 'some-toolversion',
                'configuration': {},
            }
        }

        actual_mimetype = converters.db_to_mimetype(input_mimetype)

        self.assertEqual(actual_mimetype, expected_mimetype)

    def test_db_to_language(self):
        input_language = {
            'id': b'some-id',
            'tool_id': 20,
            'tool_name': 'some-toolname',
            'tool_version': 'some-toolversion',
            'tool_configuration': {},
            'lang': b'css',
        }

        expected_language = {
            'id': b'some-id',
            'lang': b'css',
            'tool': {
                'id': 20,
                'name': 'some-toolname',
                'version': 'some-toolversion',
                'configuration': {},
            }
        }

        actual_language = converters.db_to_language(input_language)

        self.assertEqual(actual_language, expected_language)

    def test_db_to_fossology_license(self):
        input_license = {
            'id': b'some-id',
            'tool_id': 20,
            'tool_name': 'nomossa',
            'tool_version': '5.22',
            'tool_configuration': {},
            'licenses': ['GPL2.0'],
        }

        expected_license = {
            'licenses': ['GPL2.0'],
            'tool': {
                'id': 20,
                'name': 'nomossa',
                'version': '5.22',
                'configuration': {},
            }
        }

        actual_license = converters.db_to_fossology_license(input_license)

        self.assertEqual(actual_license, expected_license)

    def test_db_to_metadata(self):
        input_metadata = {
            'id': b'some-id',
            'tool_id': 20,
            'tool_name': 'some-toolname',
            'tool_version': 'some-toolversion',
            'tool_configuration': {},
            'translated_metadata': b'translated_metadata',
        }

        expected_metadata = {
            'id': b'some-id',
            'translated_metadata': b'translated_metadata',
            'tool': {
                'id': 20,
                'name': 'some-toolname',
                'version': 'some-toolversion',
                'configuration': {},
            }
        }

        actual_metadata = converters.db_to_metadata(input_metadata)

        self.assertEqual(actual_metadata, expected_metadata)



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.core.tests.server_testing import ServerTestFixture
from swh.indexer.storage import INDEXER_CFG_KEY
from swh.indexer.storage.api.client import RemoteStorage
from swh.indexer.storage.api.server import app

from .test_storage import CommonTestStorage, BasePgTestStorage


class TestRemoteStorage(CommonTestStorage, ServerTestFixture,
                        BasePgTestStorage, unittest.TestCase):
    """Test the indexer's remote storage API.

    This class doesn't define any tests as we want identical
    functionality between local and remote storage. All the tests are
    therefore defined in
    `class`:swh.indexer.storage.test_storage.CommonTestStorage.

    """

    def setUp(self):
        self.config = {
            INDEXER_CFG_KEY: {
                'cls': 'local',
                'args': {
                    'db': 'dbname=%s' % self.TEST_DB_NAME,
                }
            }
        }
        self.app = app
        super().setUp()
        self.storage = RemoteStorage(self.url())



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import reduce
import tempfile
from unittest.mock import patch

from click.testing import CliRunner

from swh.model.hashutil import hash_to_bytes

from swh.indexer.cli import cli


CLI_CONFIG = '''
scheduler:
    cls: foo
    args: {}
storage:
    cls: memory
    args: {}
indexer_storage:
    cls: memory
    args: {}
'''


def fill_idx_storage(idx_storage, nb_rows):
    tools = [
        {
            'tool_name': 'tool %d' % i,
            'tool_version': '0.0.1',
            'tool_configuration': {},
        }
        for i in range(2)
    ]
    tools = idx_storage.indexer_configuration_add(tools)

    origin_metadata = [
        {
            'origin_id': origin_id,
            'from_revision': hash_to_bytes('abcd{:0>4}'.format(origin_id)),
            'indexer_configuration_id': tools[origin_id % 2]['id'],
            'metadata': {'name': 'origin %d' % origin_id},
            'mappings': ['mapping%d' % (origin_id % 10)]
        }
        for origin_id in range(nb_rows)
    ]
    revision_metadata = [
        {
            'id': hash_to_bytes('abcd{:0>4}'.format(origin_id)),
            'indexer_configuration_id': tools[origin_id % 2]['id'],
            'metadata': {'name': 'origin %d' % origin_id},
            'mappings': ['mapping%d' % (origin_id % 10)]
        }
        for origin_id in range(nb_rows)
    ]

    idx_storage.revision_metadata_add(revision_metadata)
    idx_storage.origin_intrinsic_metadata_add(origin_metadata)

    return [tool['id'] for tool in tools]


def _origins_in_task_args(tasks):
    """Returns the set of origins contained in the arguments of the
    provided tasks (assumed to be of type indexer_origin_metadata)."""
    return reduce(
        set.union,
        (set(task['arguments']['args'][0]) for task in tasks),
        set()
    )


def _assert_tasks_for_origins(tasks, origins):
    expected_kwargs = {"policy_update": "update-dups", "parse_ids": False}
    assert {task['type'] for task in tasks} == {'indexer_origin_metadata'}
    assert all(len(task['arguments']['args']) == 1 for task in tasks)
    assert all(task['arguments']['kwargs'] == expected_kwargs
               for task in tasks)
    assert _origins_in_task_args(tasks) == set(origins)


def invoke(scheduler, catch_exceptions, args):
    runner = CliRunner()
    with patch('swh.indexer.cli.get_scheduler') as get_scheduler_mock, \
            tempfile.NamedTemporaryFile('a', suffix='.yml') as config_fd:
        config_fd.write(CLI_CONFIG)
        config_fd.seek(0)
        get_scheduler_mock.return_value = scheduler
        result = runner.invoke(cli, ['-C' + config_fd.name] + args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test_mapping_list(indexer_scheduler):
    result = invoke(indexer_scheduler, False, [
        'mapping', 'list',
    ])
    expected_output = '\n'.join([
        'codemeta', 'gemspec', 'maven', 'npm', 'pkg-info', '',
    ])
    assert result.exit_code == 0, result.output
    assert result.output == expected_output


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_empty_db(
        indexer_scheduler, idx_storage, storage):
    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
    ])
    expected_output = (
        'Nothing to do (no origin metadata matched the criteria).\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 0


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_divisor(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 90)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (30 origins).\n'
        'Scheduled 6 tasks (60 origins).\n'
        'Scheduled 9 tasks (90 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 9
    _assert_tasks_for_origins(tasks, range(90))


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_dry_run(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 90)

    result = invoke(indexer_scheduler, False, [
        'schedule', '--dry-run', 'reindex_origin_metadata',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (30 origins).\n'
        'Scheduled 6 tasks (60 origins).\n'
        'Scheduled 9 tasks (90 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 0


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_nondivisor(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when neither origin_batch_size or
    task_batch_size is a divisor of nb_origins."""
    fill_idx_storage(idx_storage, 70)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--batch-size', '20',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (60 origins).\n'
        'Scheduled 4 tasks (70 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 4
    _assert_tasks_for_origins(tasks, range(70))


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_filter_one_mapping(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 110)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--mapping', 'mapping1',
    ])

    # Check the output
    expected_output = (
        'Scheduled 2 tasks (11 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 2
    _assert_tasks_for_origins(
        tasks,
        [1, 11, 21, 31, 41, 51, 61, 71, 81, 91, 101])


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_filter_two_mappings(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    fill_idx_storage(idx_storage, 110)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--mapping', 'mapping1', '--mapping', 'mapping2',
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (22 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 3
    _assert_tasks_for_origins(
        tasks,
        [1, 11, 21, 31, 41, 51, 61, 71, 81, 91, 101,
         2, 12, 22, 32, 42, 52, 62, 72, 82, 92, 102])


@patch('swh.indexer.cli.TASK_BATCH_SIZE', 3)
def test_origin_metadata_reindex_filter_one_tool(
        indexer_scheduler, idx_storage, storage):
    """Tests the re-indexing when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    tool_ids = fill_idx_storage(idx_storage, 110)

    result = invoke(indexer_scheduler, False, [
        'schedule', 'reindex_origin_metadata',
        '--tool-id', str(tool_ids[0]),
    ])

    # Check the output
    expected_output = (
        'Scheduled 3 tasks (30 origins).\n'
        'Scheduled 6 tasks (55 origins).\n'
        'Done.\n'
    )
    assert result.exit_code == 0, result.output
    assert result.output == expected_output

    # Check scheduled tasks
    tasks = indexer_scheduler.search_tasks()
    assert len(tasks) == 6
    _assert_tasks_for_origins(
        tasks,
        [x*2 for x in range(55)])



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import ast
import os
import logging
import shutil
import tempfile
import datetime
from copy import deepcopy
from contextlib import contextmanager

from swh.scheduler import get_scheduler
try:
    from swh.scheduler import CONFIG as SWH_CONFIG
except ImportError:
    # for swh-scheduler < 0.0.47 bw compat
    SWH_CONFIG = None

from swh.storage import get_storage
from swh.core.config import SWHConfig
from swh.objstorage import get_objstorage
from swh.objstorage.exc import ObjNotFoundError
from swh.indexer.storage import get_indexer_storage, INDEXER_CFG_KEY
from swh.model import hashutil
from swh.core import utils


@contextmanager
def write_to_temp(filename, data, working_directory):
    """Write the sha1's content in a temporary file.

    Args:
        filename (str): one of sha1's many filenames
        data (bytes): the sha1's content to write in temporary
          file

    Returns:
        The path to the temporary file created. That file is
        filled in with the raw content's data.

    """
    os.makedirs(working_directory, exist_ok=True)
    temp_dir = tempfile.mkdtemp(dir=working_directory)
    content_path = os.path.join(temp_dir, filename)

    with open(content_path, 'wb') as f:
        f.write(data)

    yield content_path
    shutil.rmtree(temp_dir)


class BaseIndexer(SWHConfig, metaclass=abc.ABCMeta):
    """Base class for indexers to inherit from.

    The main entry point is the :func:`run` function which is in
    charge of triggering the computations on the batch dict/ids
    received.

    Indexers can:

    - filter out ids whose data has already been indexed.
    - retrieve ids data from storage or objstorage
    - index this data depending on the object and store the result in
      storage.

    To implement a new object type indexer, inherit from the
    BaseIndexer and implement indexing:

    :meth:`~BaseIndexer.run`:
      object_ids are different depending on object. For example: sha1 for
      content, sha1_git for revision, directory, release, and id for origin

    To implement a new concrete indexer, inherit from the object level
    classes: :class:`ContentIndexer`, :class:`RevisionIndexer`,
    :class:`OriginIndexer`.

    Then you need to implement the following functions:

    :meth:`~BaseIndexer.filter`:
      filter out data already indexed (in storage).

    :meth:`~BaseIndexer.index_object`:
      compute index on id with data (retrieved from the storage or the
      objstorage by the id key) and return the resulting index computation.

    :meth:`~BaseIndexer.persist_index_computations`:
      persist the results of multiple index computations in the storage.

    The new indexer implementation can also override the following functions:

    :meth:`~BaseIndexer.prepare`:
      Configuration preparation for the indexer.  When overriding, this must
      call the `super().prepare()` instruction.

    :meth:`~BaseIndexer.check`:
      Configuration check for the indexer.  When overriding, this must call the
      `super().check()` instruction.

    :meth:`~BaseIndexer.register_tools`:
      This should return a dict of the tool(s) to use when indexing or
      filtering.

    """
    CONFIG = 'indexer/base'

    DEFAULT_CONFIG = {
        INDEXER_CFG_KEY: ('dict', {
            'cls': 'remote',
            'args': {
                'url': 'http://localhost:5007/'
            }
        }),
        'storage': ('dict', {
            'cls': 'remote',
            'args': {
                'url': 'http://localhost:5002/',
            }
        }),
        'objstorage': ('dict', {
            'cls': 'remote',
            'args': {
                'url': 'http://localhost:5003/',
            }
        })
    }

    ADDITIONAL_CONFIG = {}

    USE_TOOLS = True

    def __init__(self, config=None, **kw):
        """Prepare and check that the indexer is ready to run.

        """
        super().__init__()
        if config is not None:
            self.config = config
        elif SWH_CONFIG:
            self.config = SWH_CONFIG.copy()
        else:
            config_keys = ('base_filename', 'config_filename',
                           'additional_configs', 'global_config')
            config_args = {k: v for k, v in kw.items() if k in config_keys}
            if self.ADDITIONAL_CONFIG:
                config_args.setdefault('additional_configs', []).append(
                    self.ADDITIONAL_CONFIG)
            self.config = self.parse_config_file(**config_args)
        self.prepare()
        self.check()
        self.log.debug('%s: config=%s', self, self.config)

    def prepare(self):
        """Prepare the indexer's needed runtime configuration.
           Without this step, the indexer cannot possibly run.

        """
        config_storage = self.config.get('storage')
        if config_storage:
            self.storage = get_storage(**config_storage)

        objstorage = self.config['objstorage']
        self.objstorage = get_objstorage(objstorage['cls'],
                                         objstorage['args'])

        idx_storage = self.config[INDEXER_CFG_KEY]
        self.idx_storage = get_indexer_storage(**idx_storage)

        _log = logging.getLogger('requests.packages.urllib3.connectionpool')
        _log.setLevel(logging.WARN)
        self.log = logging.getLogger('swh.indexer')

        if self.USE_TOOLS:
            self.tools = list(self.register_tools(
                self.config.get('tools', [])))
        self.results = []

    @property
    def tool(self):
        return self.tools[0]

    def check(self):
        """Check the indexer's configuration is ok before proceeding.
           If ok, does nothing. If not raise error.

        """
        if self.USE_TOOLS and not self.tools:
            raise ValueError('Tools %s is unknown, cannot continue' %
                             self.tools)

    def _prepare_tool(self, tool):
        """Prepare the tool dict to be compliant with the storage api.

        """
        return {'tool_%s' % key: value for key, value in tool.items()}

    def register_tools(self, tools):
        """Permit to register tools to the storage.

           Add a sensible default which can be overridden if not
           sufficient.  (For now, all indexers use only one tool)

           Expects the self.config['tools'] property to be set with
           one or more tools.

        Args:
            tools (dict/[dict]): Either a dict or a list of dict.

        Returns:
            list: List of dicts with additional id key.

        Raises:
            ValueError: if not a list nor a dict.

        """
        if isinstance(tools, list):
            tools = list(map(self._prepare_tool, tools))
        elif isinstance(tools, dict):
            tools = [self._prepare_tool(tools)]
        else:
            raise ValueError('Configuration tool(s) must be a dict or list!')

        if tools:
            return self.idx_storage.indexer_configuration_add(tools)
        else:
            return []

    def index(self, id, data):
        """Index computation for the id and associated raw data.

        Args:
            id (bytes): identifier
            data (bytes): id's data from storage or objstorage depending on
              object type

        Returns:
            dict: a dict that makes sense for the
            :meth:`.persist_index_computations` method.

        """
        raise NotImplementedError()

    def filter(self, ids):
        """Filter missing ids for that particular indexer.

        Args:
            ids ([bytes]): list of ids

        Yields:
            iterator of missing ids

        """
        yield from ids

    @abc.abstractmethod
    def persist_index_computations(self, results, policy_update):
        """Persist the computation resulting from the index.

        Args:

            results ([result]): List of results. One result is the
              result of the index function.
            policy_update ([str]): either 'update-dups' or 'ignore-dups' to
              respectively update duplicates or ignore them

        Returns:
            None

        """
        pass

    def next_step(self, results, task):
        """Do something else with computations results (e.g. send to another
        queue, ...).

        (This is not an abstractmethod since it is optional).

        Args:
            results ([result]): List of results (dict) as returned
              by index function.
            task (dict): a dict in the form expected by
              `scheduler.backend.SchedulerBackend.create_tasks`
              without `next_run`, plus an optional `result_name` key.

        Returns:
            None

        """
        if task:
            if getattr(self, 'scheduler', None):
                scheduler = self.scheduler
            else:
                scheduler = get_scheduler(**self.config['scheduler'])
            task = deepcopy(task)
            result_name = task.pop('result_name', None)
            task['next_run'] = datetime.datetime.now()
            if result_name:
                task['arguments']['kwargs'][result_name] = self.results
            scheduler.create_tasks([task])

    @abc.abstractmethod
    def run(self, ids, policy_update,
            next_step=None, **kwargs):
        """Given a list of ids:

        - retrieves the data from the storage
        - executes the indexing computations
        - stores the results (according to policy_update)

        Args:
            ids ([bytes]): id's identifier list
            policy_update (str): either 'update-dups' or 'ignore-dups' to
              respectively update duplicates or ignore them
            next_step (dict): a dict in the form expected by
              `scheduler.backend.SchedulerBackend.create_tasks`
              without `next_run`, plus a `result_name` key.
            **kwargs: passed to the `index` method

        """
        pass


class ContentIndexer(BaseIndexer):
    """A content indexer working on a list of ids directly.

    To work on indexer range, use the :class:`ContentRangeIndexer`
    instead.

    Note: :class:`ContentIndexer` is not an instantiable object. To
    use it, one should inherit from this class and override the
    methods mentioned in the :class:`BaseIndexer` class.

    """

    def run(self, ids, policy_update,
            next_step=None, **kwargs):
        """Given a list of ids:

        - retrieve the content from the storage
        - execute the indexing computations
        - store the results (according to policy_update)

        Args:
            ids (Iterable[Union[bytes, str]]): sha1's identifier list
            policy_update (str): either 'update-dups' or 'ignore-dups' to
                                 respectively update duplicates or ignore
                                 them
            next_step (dict): a dict in the form expected by
                        `scheduler.backend.SchedulerBackend.create_tasks`
                        without `next_run`, plus an optional `result_name` key.
            **kwargs: passed to the `index` method

        """
        ids = [hashutil.hash_to_bytes(id_) if isinstance(id_, str) else id_
               for id_ in ids]
        results = []
        try:
            for sha1 in ids:
                try:
                    raw_content = self.objstorage.get(sha1)
                except ObjNotFoundError:
                    self.log.warning('Content %s not found in objstorage' %
                                     hashutil.hash_to_hex(sha1))
                    continue
                res = self.index(sha1, raw_content, **kwargs)
                if res:  # If no results, skip it
                    results.append(res)

            self.persist_index_computations(results, policy_update)
            self.results = results
            return self.next_step(results, task=next_step)
        except Exception:
            self.log.exception(
                'Problem when reading contents metadata.')


class ContentRangeIndexer(BaseIndexer):
    """A content range indexer.

    This expects as input a range of ids to index.

    To work on a list of ids, use the :class:`ContentIndexer` instead.

    Note: :class:`ContentRangeIndexer` is not an instantiable
    object. To use it, one should inherit from this class and override
    the methods mentioned in the :class:`BaseIndexer` class.

    """
    @abc.abstractmethod
    def indexed_contents_in_range(self, start, end):
        """Retrieve indexed contents within range [start, end].

        Args:
            start (bytes): Starting bound from range identifier
            end (bytes): End range identifier

        Yields:
            bytes: Content identifier present in the range ``[start, end]``

        """
        pass

    def _list_contents_to_index(self, start, end, indexed):
        """Compute from storage the new contents to index in the range [start,
           end]. The already indexed contents are skipped.

        Args:
            start (bytes): Starting bound from range identifier
            end (bytes): End range identifier
            indexed (Set[bytes]): Set of content already indexed.

        Yields:
            bytes: Identifier of contents to index.

        """
        if not isinstance(start, bytes) or not isinstance(end, bytes):
            raise TypeError('identifiers must be bytes, not %r and %r.' %
                            (start, end))
        while start:
            result = self.storage.content_get_range(start, end)
            contents = result['contents']
            for c in contents:
                _id = hashutil.hash_to_bytes(c['sha1'])
                if _id in indexed:
                    continue
                yield _id
            start = result['next']

    def _index_contents(self, start, end, indexed, **kwargs):
        """Index the contents from within range [start, end]

        Args:
            start (bytes): Starting bound from range identifier
            end (bytes): End range identifier
            indexed (Set[bytes]): Set of content already indexed.

        Yields:
            dict: Data indexed to persist using the indexer storage

        """
        for sha1 in self._list_contents_to_index(start, end, indexed):
            try:
                raw_content = self.objstorage.get(sha1)
            except ObjNotFoundError:
                self.log.warning('Content %s not found in objstorage' %
                                 hashutil.hash_to_hex(sha1))
                continue
            res = self.index(sha1, raw_content, **kwargs)
            if res:
                if not isinstance(res['id'], bytes):
                    raise TypeError(
                        '%r.index should return ids as bytes, not %r' %
                        (self.__class__.__name__, res['id']))
                yield res

    def _index_with_skipping_already_done(self, start, end):
        """Index not already indexed contents in range [start, end].

        Args:
            start** (Union[bytes, str]): Starting range identifier
            end (Union[bytes, str]): Ending range identifier

        Yields:
            bytes: Content identifier present in the range
            ``[start, end]`` which are not already indexed.

        """
        while start:
            indexed_page = self.indexed_contents_in_range(start, end)
            contents = indexed_page['ids']
            _end = contents[-1] if contents else end
            yield from self._index_contents(
                    start, _end, contents)
            start = indexed_page['next']

    def run(self, start, end, skip_existing=True, **kwargs):
        """Given a range of content ids, compute the indexing computations on
           the contents within. Either the indexer is incremental
           (filter out existing computed data) or not (compute
           everything from scratch).

        Args:
            start (Union[bytes, str]): Starting range identifier
            end (Union[bytes, str]): Ending range identifier
            skip_existing (bool): Skip existing indexed data
              (default) or not
            **kwargs: passed to the `index` method

        Returns:
            bool: True if data was indexed, False otherwise.

        """
        with_indexed_data = False
        try:
            if isinstance(start, str):
                start = hashutil.hash_to_bytes(start)
            if isinstance(end, str):
                end = hashutil.hash_to_bytes(end)

            if skip_existing:
                gen = self._index_with_skipping_already_done(start, end)
            else:
                gen = self._index_contents(start, end, indexed=[])

            for results in utils.grouper(gen,
                                         n=self.config['write_batch_size']):
                self.persist_index_computations(
                    results, policy_update='update-dups')
                with_indexed_data = True
        except Exception:
            self.log.exception(
                'Problem when computing metadata.')
        finally:
            return with_indexed_data


def origin_get_params(id_):
    """From any of the two types of origin identifiers (int or
    type+url), returns a dict that can be passed to Storage.origin_get.
    Also accepts JSON-encoded forms of these (used via the task scheduler).

    >>> from pprint import pprint
    >>> origin_get_params(123)
    {'id': 123}
    >>> pprint(origin_get_params(['git', 'https://example.com/foo.git']))
    {'type': 'git', 'url': 'https://example.com/foo.git'}
    >>> origin_get_params("123")
    {'id': 123}
    >>> pprint(origin_get_params('["git", "https://example.com/foo.git"]'))
    {'type': 'git', 'url': 'https://example.com/foo.git'}
    """
    if isinstance(id_, str):
        # Data coming from JSON, which requires string keys, so
        # one extra level of deserialization is needed
        id_ = ast.literal_eval(id_)
    if isinstance(id_, (tuple, list)):
        if len(id_) != 2:
            raise TypeError('Expected a (type, url) tuple.')
        (type_, url) = id_
        params = {'type': type_, 'url': url}
    elif isinstance(id_, int):
        params = {'id': id_}
    else:
        raise TypeError('Invalid value in "ids": %r' % id_)
    return params


class OriginIndexer(BaseIndexer):
    """An object type indexer, inherits from the :class:`BaseIndexer` and
    implements Origin indexing using the run method

    Note: the :class:`OriginIndexer` is not an instantiable object.
    To use it in another context one should inherit from this class
    and override the methods mentioned in the :class:`BaseIndexer`
    class.

    """
    def run(self, ids, policy_update='update-dups', parse_ids=True,
            next_step=None, **kwargs):
        """Given a list of origin ids:

        - retrieve origins from storage
        - execute the indexing computations
        - store the results (according to policy_update)

        Args:
            ids ([Union[int, Tuple[str, bytes]]]): list of origin ids or
              (type, url) tuples.
            policy_update (str): either 'update-dups' or 'ignore-dups' to
              respectively update duplicates (default) or ignore them
            next_step (dict): a dict in the form expected by
              `scheduler.backend.SchedulerBackend.create_tasks` without
              `next_run`, plus an optional `result_name` key.
            parse_ids (bool): Do we need to parse id or not (default)
            **kwargs: passed to the `index` method

        """
        if parse_ids:
            ids = [o.split('+', 1) if ':' in o else int(o)  # type+url or id
                   for o in ids]

        origins_filtered = []
        origins = self.storage.origin_get(
            [origin_get_params(id_) for id_ in ids])
        for (id_, origin) in zip(ids, origins):
            if not origin:
                self.log.warning('Origin %s not found in storage' %
                                 id_)
                continue
            origins_filtered.append(origin)

        results = self.index_list(origins_filtered, **kwargs)

        self.persist_index_computations(results, policy_update)
        self.results = results
        return self.next_step(results, task=next_step)

    def index_list(self, origins, **kwargs):
        results = []
        for origin in origins:
            try:
                res = self.index(origin, **kwargs)
                if res:  # If no results, skip it
                    results.append(res)
            except Exception:
                self.log.exception(
                    'Problem when processing origin %s',
                    origin['id'])
        return results


class RevisionIndexer(BaseIndexer):
    """An object type indexer, inherits from the :class:`BaseIndexer` and
    implements Revision indexing using the run method

    Note: the :class:`RevisionIndexer` is not an instantiable object.
    To use it in another context one should inherit from this class
    and override the methods mentioned in the :class:`BaseIndexer`
    class.

    """
    def run(self, ids, policy_update, next_step=None):
        """Given a list of sha1_gits:

        - retrieve revisions from storage
        - execute the indexing computations
        - store the results (according to policy_update)

        Args:
            ids ([bytes or str]): sha1_git's identifier list
            policy_update (str): either 'update-dups' or 'ignore-dups' to
              respectively update duplicates or ignore them

        """
        results = []
        ids = [hashutil.hash_to_bytes(id_) if isinstance(id_, str) else id_
               for id_ in ids]
        revs = self.storage.revision_get(ids)

        for rev in revs:
            if not rev:
                self.log.warning('Revisions %s not found in storage' %
                                 list(map(hashutil.hash_to_hex, ids)))
                continue
            try:
                res = self.index(rev)
                if res:  # If no results, skip it
                    results.append(res)
            except Exception:
                self.log.exception(
                        'Problem when processing revision')
        self.persist_index_computations(results, policy_update)
        self.results = results
        return self.next_step(results, task=next_step)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click

from swh.core import config
from swh.scheduler import get_scheduler
from swh.scheduler.utils import create_task_dict
from swh.storage import get_storage

from swh.indexer.metadata_dictionary import MAPPINGS
from swh.indexer.storage import get_indexer_storage
from swh.indexer.storage.api.server import load_and_check_config, app


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

TASK_BATCH_SIZE = 1000  # Number of tasks per query to the scheduler


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--config-file', '-C', default=None,
              type=click.Path(exists=True, dir_okay=False,),
              help="Configuration file.")
@click.pass_context
def cli(ctx, config_file):
    """Software Heritage Indexer CLI interface
    """
    ctx.ensure_object(dict)

    conf = config.read(config_file)
    ctx.obj['config'] = conf


def _get_api(getter, config, config_key, url):
    if url:
        config[config_key] = {
            'cls': 'remote',
            'args': {'url': url}
        }
    elif config_key not in config:
        raise click.ClickException(
            'Missing configuration for {}'.format(config_key))
    return getter(**config[config_key])


@cli.group('mapping')
def mapping():
    pass


@mapping.command('list')
def mapping_list():
    """Prints the list of known mappings."""
    mapping_names = [mapping.name for mapping in MAPPINGS.values()]
    mapping_names.sort()
    for mapping_name in mapping_names:
        click.echo(mapping_name)


@cli.group('schedule')
@click.option('--scheduler-url', '-s', default=None,
              help="URL of the scheduler API")
@click.option('--indexer-storage-url', '-i', default=None,
              help="URL of the indexer storage API")
@click.option('--storage-url', '-g', default=None,
              help="URL of the (graph) storage API")
@click.option('--dry-run/--no-dry-run', is_flag=True,
              default=False,
              help='Default to list only what would be scheduled.')
@click.pass_context
def schedule(ctx, scheduler_url, storage_url, indexer_storage_url,
             dry_run):
    """Manipulate indexer tasks via SWH Scheduler's API."""
    ctx.obj['indexer_storage'] = _get_api(
        get_indexer_storage,
        ctx.obj['config'],
        'indexer_storage',
        indexer_storage_url
    )
    ctx.obj['storage'] = _get_api(
        get_storage,
        ctx.obj['config'],
        'storage',
        storage_url
    )
    ctx.obj['scheduler'] = _get_api(
        get_scheduler,
        ctx.obj['config'],
        'scheduler',
        scheduler_url
    )
    if dry_run:
        ctx.obj['scheduler'] = None


def list_origins_by_producer(idx_storage, mappings, tool_ids):
    start = 0
    limit = 10000
    while True:
        origins = list(
            idx_storage.origin_intrinsic_metadata_search_by_producer(
                start=start, limit=limit, ids_only=True,
                mappings=mappings or None, tool_ids=tool_ids or None))
        if not origins:
            break
        start = origins[-1]+1
        yield from origins


@schedule.command('reindex_origin_metadata')
@click.option('--batch-size', '-b', 'origin_batch_size',
              default=10, show_default=True, type=int,
              help="Number of origins per task")
@click.option('--tool-id', '-t', 'tool_ids', type=int, multiple=True,
              help="Restrict search of old metadata to this/these tool ids.")
@click.option('--mapping', '-m', 'mappings', multiple=True,
              help="Mapping(s) that should be re-scheduled (eg. 'npm', "
                   "'gemspec', 'maven')")
@click.option('--task-type',
              default='indexer_origin_metadata', show_default=True,
              help="Name of the task type to schedule.")
@click.pass_context
def schedule_origin_metadata_reindex(
        ctx, origin_batch_size, mappings, tool_ids, task_type):
    """Schedules indexing tasks for origins that were already indexed."""
    idx_storage = ctx.obj['indexer_storage']
    scheduler = ctx.obj['scheduler']

    origins = list_origins_by_producer(idx_storage, mappings, tool_ids)
    kwargs = {"policy_update": "update-dups", "parse_ids": False}
    nb_origins = 0
    nb_tasks = 0

    while True:
        task_batch = []
        for _ in range(TASK_BATCH_SIZE):
            # Group origins
            origin_batch = []
            for (_, origin) in zip(range(origin_batch_size), origins):
                origin_batch.append(origin)
            nb_origins += len(origin_batch)
            if not origin_batch:
                break

            # Create a task for these origins
            args = [origin_batch]
            task_dict = create_task_dict(task_type, 'oneshot', *args, **kwargs)
            task_batch.append(task_dict)

        # Schedule a batch of tasks
        if not task_batch:
            break
        nb_tasks += len(task_batch)
        if scheduler:
            scheduler.create_tasks(task_batch)
        click.echo('Scheduled %d tasks (%d origins).' % (nb_tasks, nb_origins))

    # Print final status.
    if nb_tasks:
        click.echo('Done.')
    else:
        click.echo('Nothing to do (no origin metadata matched the criteria).')


@cli.command('api-server')
@click.argument('config-path', required=1)
@click.option('--host', default='0.0.0.0', help="Host to run the server")
@click.option('--port', default=5007, type=click.INT,
              help="Binding port of the server")
@click.option('--debug/--nodebug', default=True,
              help="Indicates if the server should run in debug mode")
def api_server(config_path, host, port, debug):
    api_cfg = load_and_check_config(config_path, type='any')
    app.config.update(api_cfg)
    app.run(host, port=int(port), debug=bool(debug))


def main():
    return cli(auto_envvar_prefix='SWH_INDEXER')


if __name__ == '__main__':
    main()



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from celery import current_app as app

from .mimetype import MimetypeIndexer, MimetypeRangeIndexer
from .language import LanguageIndexer
from .ctags import CtagsIndexer
from .fossology_license import (
    FossologyLicenseIndexer, FossologyLicenseRangeIndexer
)
from .rehash import RecomputeChecksums
from .metadata import (
    RevisionMetadataIndexer, OriginMetadataIndexer
)
from .origin_head import OriginHeadIndexer


@app.task(name=__name__ + '.RevisionMetadata')
def revision_metadata(*args, **kwargs):
    results = RevisionMetadataIndexer().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.OriginMetadata')
def origin_metadata(*args, **kwargs):
    results = OriginMetadataIndexer().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.OriginHead')
def origin_head(*args, **kwargs):
    results = OriginHeadIndexer().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.ContentLanguage')
def content_language(*args, **kwargs):
    results = LanguageIndexer().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.Ctags')
def ctags(*args, **kwargs):
    results = CtagsIndexer().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.ContentFossologyLicense')
def fossology_license(*args, **kwargs):
    results = FossologyLicenseIndexer().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.RecomputeChecksums')
def recompute_checksums(*args, **kwargs):
    results = RecomputeChecksums().run(*args, **kwargs)
    return getattr(results, 'results', results)


@app.task(name=__name__ + '.ContentMimetype')
def mimetype(*args, **kwargs):
    results = MimetypeIndexer().run(*args, **kwargs)
    return {'status': 'eventful' if results else 'uneventful'}


@app.task(name=__name__ + '.ContentRangeMimetype')
def range_mimetype(*args, **kwargs):
    results = MimetypeRangeIndexer().run(*args, **kwargs)
    return {'status': 'eventful' if results else 'uneventful'}


@app.task(name=__name__ + '.ContentRangeFossologyLicense')
def range_license(*args, **kwargs):
    results = FossologyLicenseRangeIndexer().run(*args, **kwargs)
    return {'status': 'eventful' if results else 'uneventful'}



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import pytest
import unittest

from hypothesis import given

from swh.model.hashutil import hash_to_bytes

from swh.indexer.storage import get_indexer_storage, MAPPING_NAMES
from swh.core.tests.db_testing import SingleDbTestFixture
from swh.indexer.tests.storage.generate_data_test import (
    gen_content_mimetypes, gen_content_fossology_licenses
)
from swh.indexer.tests.storage import SQL_DIR
from swh.indexer.metadata_dictionary import MAPPINGS

TOOLS = [
    {
        'tool_name': 'universal-ctags',
        'tool_version': '~git7859817b',
        'tool_configuration': {
            "command_line": "ctags --fields=+lnz --sort=no --links=no "
                            "--output-format=json <filepath>"}
    },
    {
        'tool_name': 'swh-metadata-translator',
        'tool_version': '0.0.1',
        'tool_configuration': {"type": "local", "context": "NpmMapping"},
    },
    {
        'tool_name': 'swh-metadata-detector',
        'tool_version': '0.0.1',
        'tool_configuration': {
            "type": "local", "context": ["NpmMapping", "CodemetaMapping"]},
    },
    {
        'tool_name': 'swh-metadata-detector2',
        'tool_version': '0.0.1',
        'tool_configuration': {
            "type": "local", "context": ["NpmMapping", "CodemetaMapping"]},
    },
    {
        'tool_name': 'file',
        'tool_version': '5.22',
        'tool_configuration': {"command_line": "file --mime <filepath>"},
    },
    {
        'tool_name': 'pygments',
        'tool_version': '2.0.1+dfsg-1.1+deb8u1',
        'tool_configuration': {
            "type": "library", "debian-package": "python3-pygments"},
    },
    {
        'tool_name': 'pygments',
        'tool_version': '2.0.1+dfsg-1.1+deb8u1',
        'tool_configuration': {
            "type": "library",
            "debian-package": "python3-pygments",
            "max_content_size": 10240
        },
    },
    {
        'tool_name': 'nomos',
        'tool_version': '3.1.0rc2-31-ga2cbb8c',
        'tool_configuration': {"command_line": "nomossa <filepath>"},
    }
]


@pytest.mark.db
class BasePgTestStorage(SingleDbTestFixture):
    """Base test class for most indexer tests.

    It adds support for Storage testing to the SingleDbTestFixture class.
    It will also build the database from the swh-indexed/sql/*.sql files.
    """

    TEST_DB_NAME = 'softwareheritage-test-indexer'
    TEST_DB_DUMP = os.path.join(SQL_DIR, '*.sql')

    def setUp(self):
        super().setUp()
        self.storage_config = {
            'cls': 'local',
            'args': {
                'db': 'dbname=%s' % self.TEST_DB_NAME,
            },
        }

    def tearDown(self):
        self.reset_storage_tables()
        self.storage = None
        super().tearDown()

    def reset_storage_tables(self):
        excluded = {'indexer_configuration'}
        self.reset_db_tables(self.TEST_DB_NAME, excluded=excluded)

        db = self.test_db[self.TEST_DB_NAME]
        db.conn.commit()


def gen_generic_endpoint_tests(endpoint_type, tool_name,
                               example_data1, example_data2):
    def rename(f):
        f.__name__ = 'test_' + endpoint_type + f.__name__
        return f

    def endpoint(self, endpoint_name):
        return getattr(self.storage, endpoint_type + '_' + endpoint_name)

    @rename
    def missing(self):
        # given
        tool_id = self.tools[tool_name]['id']

        query = [
            {
                'id': self.sha1_1,
                'indexer_configuration_id': tool_id,
            },
            {
                'id': self.sha1_2,
                'indexer_configuration_id': tool_id,
            }]

        # when
        actual_missing = endpoint(self, 'missing')(query)

        # then
        self.assertEqual(list(actual_missing), [
            self.sha1_1,
            self.sha1_2,
        ])

        # given
        endpoint(self, 'add')([{
            'id': self.sha1_2,
            **example_data1,
            'indexer_configuration_id': tool_id,
        }])

        # when
        actual_missing = endpoint(self, 'missing')(query)

        # then
        self.assertEqual(list(actual_missing), [self.sha1_1])

    @rename
    def add__drop_duplicate(self):
        # given
        tool_id = self.tools[tool_name]['id']

        data_v1 = {
            'id': self.sha1_2,
            **example_data1,
            'indexer_configuration_id': tool_id,
        }

        # given
        endpoint(self, 'add')([data_v1])

        # when
        actual_data = list(endpoint(self, 'get')([self.sha1_2]))

        # then
        expected_data_v1 = [{
            'id': self.sha1_2,
            **example_data1,
            'tool': self.tools[tool_name],
        }]
        self.assertEqual(actual_data, expected_data_v1)

        # given
        data_v2 = data_v1.copy()
        data_v2.update(example_data2)

        endpoint(self, 'add')([data_v2])

        actual_data = list(endpoint(self, 'get')([self.sha1_2]))

        # data did not change as the v2 was dropped.
        self.assertEqual(actual_data, expected_data_v1)

    @rename
    def add__update_in_place_duplicate(self):
        # given
        tool_id = self.tools[tool_name]['id']

        data_v1 = {
            'id': self.sha1_2,
            **example_data1,
            'indexer_configuration_id': tool_id,
        }

        # given
        endpoint(self, 'add')([data_v1])

        # when
        actual_data = list(endpoint(self, 'get')([self.sha1_2]))

        expected_data_v1 = [{
            'id': self.sha1_2,
            **example_data1,
            'tool': self.tools[tool_name],
        }]

        # then
        self.assertEqual(actual_data, expected_data_v1)

        # given
        data_v2 = data_v1.copy()
        data_v2.update(example_data2)

        endpoint(self, 'add')([data_v2], conflict_update=True)

        actual_data = list(endpoint(self, 'get')([self.sha1_2]))

        expected_data_v2 = [{
            'id': self.sha1_2,
            **example_data2,
            'tool': self.tools[tool_name],
        }]

        # data did change as the v2 was used to overwrite v1
        self.assertEqual(actual_data, expected_data_v2)

    def add__duplicate_twice(self):
        # given
        tool_id = self.tools[tool_name]['id']

        data_rev1 = {
            'id': self.revision_id_2,
            **example_data1,
            'indexer_configuration_id': tool_id
        }

        data_rev2 = {
            'id': self.revision_id_2,
            **example_data2,
            'indexer_configuration_id': tool_id
        }

        # when
        endpoint(self, 'add')([data_rev1])

        with self.assertRaises(ValueError):
            endpoint(self, 'add')(
                [data_rev2, data_rev2],
                conflict_update=True)

        # then
        actual_data = list(endpoint(self, 'get')(
            [self.revision_id_2, self.revision_id_1]))

        expected_data = [{
            'id': self.revision_id_2,
            **example_data1,
            'tool': self.tools[tool_name]
        }]
        self.assertEqual(actual_data, expected_data)

    @rename
    def get(self):
        # given
        tool_id = self.tools[tool_name]['id']

        query = [self.sha1_2, self.sha1_1]

        data1 = {
            'id': self.sha1_2,
            **example_data1,
            'indexer_configuration_id': tool_id,
        }

        # when
        endpoint(self, 'add')([data1])

        # then
        actual_data = list(endpoint(self, 'get')(query))

        # then
        expected_data = [{
            'id': self.sha1_2,
            **example_data1,
            'tool': self.tools[tool_name]
        }]

        self.assertEqual(actual_data, expected_data)

    return (
        missing,
        add__drop_duplicate,
        add__update_in_place_duplicate,
        add__duplicate_twice,
        get,
    )


class CommonTestStorage:
    """Base class for Indexer Storage testing.

    """
    def setUp(self):
        super().setUp()
        self.storage = get_indexer_storage(**self.storage_config)
        tools = self.storage.indexer_configuration_add(TOOLS)
        self.tools = {}
        for tool in tools:
            tool_name = tool['tool_name']
            while tool_name in self.tools:
                tool_name += '_'
            self.tools[tool_name] = {
                'id': tool['id'],
                'name': tool['tool_name'],
                'version': tool['tool_version'],
                'configuration': tool['tool_configuration'],
            }

        self.sha1_1 = hash_to_bytes('34973274ccef6ab4dfaaf86599792fa9c3fe4689')
        self.sha1_2 = hash_to_bytes('61c2b3a30496d329e21af70dd2d7e097046d07b7')
        self.revision_id_1 = hash_to_bytes(
            '7026b7c1a2af56521e951c01ed20f255fa054238')
        self.revision_id_2 = hash_to_bytes(
            '7026b7c1a2af56521e9587659012345678904321')
        self.revision_id_3 = hash_to_bytes(
            '7026b7c1a2af56521e9587659012345678904320')
        self.origin_id_1 = 44434341
        self.origin_id_2 = 44434342
        self.origin_id_3 = 54974445

    def test_check_config(self):
        self.assertTrue(self.storage.check_config(check_write=True))
        self.assertTrue(self.storage.check_config(check_write=False))

    # generate content_mimetype tests
    (
        test_content_mimetype_missing,
        test_content_mimetype_add__drop_duplicate,
        test_content_mimetype_add__update_in_place_duplicate,
        test_content_mimetype_add__duplicate_twice,
        test_content_mimetype_get,
    ) = gen_generic_endpoint_tests(
        endpoint_type='content_mimetype',
        tool_name='file',
        example_data1={
            'mimetype': 'text/plain',
            'encoding': 'utf-8',
        },
        example_data2={
            'mimetype': 'text/html',
            'encoding': 'us-ascii',
        },
    )

    # content_language tests
    (
        test_content_language_missing,
        test_content_language_add__drop_duplicate,
        test_content_language_add__update_in_place_duplicate,
        test_content_language_add__duplicate_twice,
        test_content_language_get,
    ) = gen_generic_endpoint_tests(
        endpoint_type='content_language',
        tool_name='pygments',
        example_data1={
            'lang': 'haskell',
        },
        example_data2={
            'lang': 'common-lisp',
        },
    )

    # content_ctags tests
    (
        test_content_ctags_missing,
        # the following tests are disabled because CTAGS behave differently
        _,  # test_content_ctags_add__drop_duplicate,
        _,  # test_content_ctags_add__update_in_place_duplicate,
        _,  # test_content_ctags_add__duplicate_twice,
        _,  # test_content_ctags_get,
    ) = gen_generic_endpoint_tests(
        endpoint_type='content_ctags',
        tool_name='universal-ctags',
        example_data1={
            'ctags': [{
                'name': 'done',
                'kind': 'variable',
                'line': 119,
                'lang': 'OCaml',
            }]
        },
        example_data2={
            'ctags': [
                {
                    'name': 'done',
                    'kind': 'variable',
                    'line': 100,
                    'lang': 'Python',
                },
                {
                    'name': 'main',
                    'kind': 'function',
                    'line': 119,
                    'lang': 'Python',
                }]
        },
    )

    def test_content_ctags_search(self):
        # 1. given
        tool = self.tools['universal-ctags']
        tool_id = tool['id']

        ctag1 = {
            'id': self.sha1_1,
            'indexer_configuration_id': tool_id,
            'ctags': [
                {
                    'name': 'hello',
                    'kind': 'function',
                    'line': 133,
                    'lang': 'Python',
                },
                {
                    'name': 'counter',
                    'kind': 'variable',
                    'line': 119,
                    'lang': 'Python',
                },
                {
                    'name': 'hello',
                    'kind': 'variable',
                    'line': 210,
                    'lang': 'Python',
                },
            ]
        }

        ctag2 = {
            'id': self.sha1_2,
            'indexer_configuration_id': tool_id,
            'ctags': [
                {
                    'name': 'hello',
                    'kind': 'variable',
                    'line': 100,
                    'lang': 'C',
                },
                {
                    'name': 'result',
                    'kind': 'variable',
                    'line': 120,
                    'lang': 'C',
                },
            ]
        }

        self.storage.content_ctags_add([ctag1, ctag2])

        # 1. when
        actual_ctags = list(self.storage.content_ctags_search('hello',
                                                              limit=1))

        # 1. then
        self.assertEqual(actual_ctags, [
            {
                'id': ctag1['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'function',
                'line': 133,
                'lang': 'Python',
            }
        ])

        # 2. when
        actual_ctags = list(self.storage.content_ctags_search(
            'hello',
            limit=1,
            last_sha1=ctag1['id']))

        # 2. then
        self.assertEqual(actual_ctags, [
            {
                'id': ctag2['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'variable',
                'line': 100,
                'lang': 'C',
            }
        ])

        # 3. when
        actual_ctags = list(self.storage.content_ctags_search('hello'))

        # 3. then
        self.assertEqual(actual_ctags, [
            {
                'id': ctag1['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'function',
                'line': 133,
                'lang': 'Python',
            },
            {
                'id': ctag1['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'variable',
                'line': 210,
                'lang': 'Python',
            },
            {
                'id': ctag2['id'],
                'tool': tool,
                'name': 'hello',
                'kind': 'variable',
                'line': 100,
                'lang': 'C',
            },
        ])

        # 4. when
        actual_ctags = list(self.storage.content_ctags_search('counter'))

        # then
        self.assertEqual(actual_ctags, [{
            'id': ctag1['id'],
            'tool': tool,
            'name': 'counter',
            'kind': 'variable',
            'line': 119,
            'lang': 'Python',
        }])

        # 5. when
        actual_ctags = list(self.storage.content_ctags_search('result',
                                                              limit=1))

        # then
        self.assertEqual(actual_ctags, [{
            'id': ctag2['id'],
            'tool': tool,
            'name': 'result',
            'kind': 'variable',
            'line': 120,
            'lang': 'C',
        }])

    def test_content_ctags_search_no_result(self):
        actual_ctags = list(self.storage.content_ctags_search('counter'))

        self.assertEqual(actual_ctags, [])

    def test_content_ctags_add__add_new_ctags_added(self):
        # given
        tool = self.tools['universal-ctags']
        tool_id = tool['id']

        ctag_v1 = {
            'id': self.sha1_2,
            'indexer_configuration_id': tool_id,
            'ctags': [{
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
            }]
        }

        # given
        self.storage.content_ctags_add([ctag_v1])
        self.storage.content_ctags_add([ctag_v1])  # conflict does nothing

        # when
        actual_ctags = list(self.storage.content_ctags_get(
            [self.sha1_2]))

        # then
        expected_ctags = [{
            'id': self.sha1_2,
            'name': 'done',
            'kind': 'variable',
            'line': 100,
            'lang': 'Scheme',
            'tool': tool,
        }]

        self.assertEqual(actual_ctags, expected_ctags)

        # given
        ctag_v2 = ctag_v1.copy()
        ctag_v2.update({
            'ctags': [
                {
                    'name': 'defn',
                    'kind': 'function',
                    'line': 120,
                    'lang': 'Scheme',
                }
            ]
        })

        self.storage.content_ctags_add([ctag_v2])

        expected_ctags = [
            {
                'id': self.sha1_2,
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
                'tool': tool,
            }, {
                'id': self.sha1_2,
                'name': 'defn',
                'kind': 'function',
                'line': 120,
                'lang': 'Scheme',
                'tool': tool,
            }
        ]

        actual_ctags = list(self.storage.content_ctags_get(
            [self.sha1_2]))

        self.assertEqual(actual_ctags, expected_ctags)

    def test_content_ctags_add__update_in_place(self):
        # given
        tool = self.tools['universal-ctags']
        tool_id = tool['id']

        ctag_v1 = {
            'id': self.sha1_2,
            'indexer_configuration_id': tool_id,
            'ctags': [{
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
            }]
        }

        # given
        self.storage.content_ctags_add([ctag_v1])

        # when
        actual_ctags = list(self.storage.content_ctags_get(
            [self.sha1_2]))

        # then
        expected_ctags = [
            {
                'id': self.sha1_2,
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
                'tool': tool
            }
        ]
        self.assertEqual(actual_ctags, expected_ctags)

        # given
        ctag_v2 = ctag_v1.copy()
        ctag_v2.update({
            'ctags': [
                {
                    'name': 'done',
                    'kind': 'variable',
                    'line': 100,
                    'lang': 'Scheme',
                },
                {
                    'name': 'defn',
                    'kind': 'function',
                    'line': 120,
                    'lang': 'Scheme',
                }
            ]
        })

        self.storage.content_ctags_add([ctag_v2], conflict_update=True)

        actual_ctags = list(self.storage.content_ctags_get(
            [self.sha1_2]))

        # ctag did change as the v2 was used to overwrite v1
        expected_ctags = [
            {
                'id': self.sha1_2,
                'name': 'done',
                'kind': 'variable',
                'line': 100,
                'lang': 'Scheme',
                'tool': tool,
            },
            {
                'id': self.sha1_2,
                'name': 'defn',
                'kind': 'function',
                'line': 120,
                'lang': 'Scheme',
                'tool': tool,
            }
        ]
        self.assertEqual(actual_ctags, expected_ctags)

    # content_fossology_license tests
    (
        _,  # The endpoint content_fossology_license_missing does not exist
        # the following tests are disabled because fossology_license tests
        # behave differently
        _,  # test_content_fossology_license_add__drop_duplicate,
        _,  # test_content_fossology_license_add__update_in_place_duplicate,
        _,  # test_content_metadata_add__duplicate_twice,
        _,  # test_content_fossology_license_get,
    ) = gen_generic_endpoint_tests(
        endpoint_type='content_fossology_license',
        tool_name='nomos',
        example_data1={
            'licenses': ['Apache-2.0'],
        },
        example_data2={
            'licenses': ['BSD-2-Clause'],
        },
    )

    def test_content_fossology_license_add__new_license_added(self):
        # given
        tool = self.tools['nomos']
        tool_id = tool['id']

        license_v1 = {
            'id': self.sha1_1,
            'licenses': ['Apache-2.0'],
            'indexer_configuration_id': tool_id,
        }

        # given
        self.storage.content_fossology_license_add([license_v1])
        # conflict does nothing
        self.storage.content_fossology_license_add([license_v1])

        # when
        actual_licenses = list(self.storage.content_fossology_license_get(
            [self.sha1_1]))

        # then
        expected_license = {
            self.sha1_1: [{
                'licenses': ['Apache-2.0'],
                'tool': tool,
            }]
        }
        self.assertEqual(actual_licenses, [expected_license])

        # given
        license_v2 = license_v1.copy()
        license_v2.update({
            'licenses': ['BSD-2-Clause'],
        })

        self.storage.content_fossology_license_add([license_v2])

        actual_licenses = list(self.storage.content_fossology_license_get(
            [self.sha1_1]))

        expected_license = {
            self.sha1_1: [{
                'licenses': ['Apache-2.0', 'BSD-2-Clause'],
                'tool': tool
            }]
        }

        # license did not change as the v2 was dropped.
        self.assertEqual(actual_licenses, [expected_license])

    # content_metadata tests
    (
        test_content_metadata_missing,
        test_content_metadata_add__drop_duplicate,
        test_content_metadata_add__update_in_place_duplicate,
        test_content_metadata_add__duplicate_twice,
        test_content_metadata_get,
    ) = gen_generic_endpoint_tests(
        endpoint_type='content_metadata',
        tool_name='swh-metadata-detector',
        example_data1={
            'translated_metadata': {
                'other': {},
                'codeRepository': {
                    'type': 'git',
                    'url': 'https://github.com/moranegg/metadata_test'
                },
                'description': 'Simple package.json test for indexer',
                'name': 'test_metadata',
                'version': '0.0.1'
            },
        },
        example_data2={
            'translated_metadata': {
                'other': {},
                'name': 'test_metadata',
                'version': '0.0.1'
            },
        },
    )

    # revision_metadata tests
    (
        test_revision_metadata_missing,
        test_revision_metadata_add__drop_duplicate,
        test_revision_metadata_add__update_in_place_duplicate,
        test_revision_metadata_add__duplicate_twice,
        test_revision_metadata_get,
    ) = gen_generic_endpoint_tests(
        endpoint_type='revision_metadata',
        tool_name='swh-metadata-detector',
        example_data1={
            'translated_metadata': {
                'other': {},
                'codeRepository': {
                    'type': 'git',
                    'url': 'https://github.com/moranegg/metadata_test'
                },
                'description': 'Simple package.json test for indexer',
                'name': 'test_metadata',
                'version': '0.0.1'
            },
            'mappings': ['mapping1'],
        },
        example_data2={
            'translated_metadata': {
                'other': {},
                'name': 'test_metadata',
                'version': '0.0.1'
            },
            'mappings': ['mapping2'],
        },
    )

    def test_origin_intrinsic_metadata_get(self):
        # given
        tool_id = self.tools['swh-metadata-detector']['id']

        metadata = {
            'version': None,
            'name': None,
        }
        metadata_rev = {
            'id': self.revision_id_2,
            'translated_metadata': metadata,
            'mappings': ['mapping1'],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin = {
            'origin_id': self.origin_id_1,
            'metadata': metadata,
            'indexer_configuration_id': tool_id,
            'mappings': ['mapping1'],
            'from_revision': self.revision_id_2,
            }

        # when
        self.storage.revision_metadata_add([metadata_rev])
        self.storage.origin_intrinsic_metadata_add([metadata_origin])

        # then
        actual_metadata = list(self.storage.origin_intrinsic_metadata_get(
            [self.origin_id_1, 42]))

        expected_metadata = [{
            'origin_id': self.origin_id_1,
            'metadata': metadata,
            'tool': self.tools['swh-metadata-detector'],
            'from_revision': self.revision_id_2,
            'mappings': ['mapping1'],
        }]

        self.assertEqual(actual_metadata, expected_metadata)

    def test_origin_intrinsic_metadata_add_drop_duplicate(self):
        # given
        tool_id = self.tools['swh-metadata-detector']['id']

        metadata_v1 = {
            'version': None,
            'name': None,
        }
        metadata_rev_v1 = {
            'id': self.revision_id_1,
            'translated_metadata': metadata_v1.copy(),
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin_v1 = {
            'origin_id': self.origin_id_1,
            'metadata': metadata_v1.copy(),
            'indexer_configuration_id': tool_id,
            'mappings': [],
            'from_revision': self.revision_id_1,
        }

        # given
        self.storage.revision_metadata_add([metadata_rev_v1])
        self.storage.origin_intrinsic_metadata_add([metadata_origin_v1])

        # when
        actual_metadata = list(self.storage.origin_intrinsic_metadata_get(
            [self.origin_id_1, 42]))

        expected_metadata_v1 = [{
            'origin_id': self.origin_id_1,
            'metadata': metadata_v1,
            'tool': self.tools['swh-metadata-detector'],
            'from_revision': self.revision_id_1,
            'mappings': [],
        }]

        self.assertEqual(actual_metadata, expected_metadata_v1)

        # given
        metadata_v2 = metadata_v1.copy()
        metadata_v2.update({
            'name': 'test_metadata',
            'author': 'MG',
        })
        metadata_rev_v2 = metadata_rev_v1.copy()
        metadata_origin_v2 = metadata_origin_v1.copy()
        metadata_rev_v2['translated_metadata'] = metadata_v2
        metadata_origin_v2['translated_metadata'] = metadata_v2

        self.storage.revision_metadata_add([metadata_rev_v2])
        self.storage.origin_intrinsic_metadata_add([metadata_origin_v2])

        # then
        actual_metadata = list(self.storage.origin_intrinsic_metadata_get(
            [self.origin_id_1]))

        # metadata did not change as the v2 was dropped.
        self.assertEqual(actual_metadata, expected_metadata_v1)

    def test_origin_intrinsic_metadata_add_update_in_place_duplicate(self):
        # given
        tool_id = self.tools['swh-metadata-detector']['id']

        metadata_v1 = {
            'version': None,
            'name': None,
        }
        metadata_rev_v1 = {
            'id': self.revision_id_2,
            'translated_metadata': metadata_v1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin_v1 = {
            'origin_id': self.origin_id_1,
            'metadata': metadata_v1.copy(),
            'indexer_configuration_id': tool_id,
            'mappings': [],
            'from_revision': self.revision_id_2,
        }

        # given
        self.storage.revision_metadata_add([metadata_rev_v1])
        self.storage.origin_intrinsic_metadata_add([metadata_origin_v1])

        # when
        actual_metadata = list(self.storage.origin_intrinsic_metadata_get(
            [self.origin_id_1]))

        # then
        expected_metadata_v1 = [{
            'origin_id': self.origin_id_1,
            'metadata': metadata_v1,
            'tool': self.tools['swh-metadata-detector'],
            'from_revision': self.revision_id_2,
            'mappings': [],
        }]
        self.assertEqual(actual_metadata, expected_metadata_v1)

        # given
        metadata_v2 = metadata_v1.copy()
        metadata_v2.update({
            'name': 'test_update_duplicated_metadata',
            'author': 'MG',
        })
        metadata_rev_v2 = metadata_rev_v1.copy()
        metadata_origin_v2 = metadata_origin_v1.copy()
        metadata_rev_v2['translated_metadata'] = metadata_v2
        metadata_origin_v2['metadata'] = metadata_v2

        self.storage.revision_metadata_add([metadata_rev_v2],
                                           conflict_update=True)
        self.storage.origin_intrinsic_metadata_add([metadata_origin_v2],
                                                   conflict_update=True)

        actual_metadata = list(self.storage.origin_intrinsic_metadata_get(
            [self.origin_id_1]))

        expected_metadata_v2 = [{
            'origin_id': self.origin_id_1,
            'metadata': metadata_v2,
            'tool': self.tools['swh-metadata-detector'],
            'from_revision': self.revision_id_2,
            'mappings': [],
        }]

        # metadata did change as the v2 was used to overwrite v1
        self.assertEqual(actual_metadata, expected_metadata_v2)

    def test_origin_intrinsic_metadata_add__duplicate_twice(self):
        # given
        tool_id = self.tools['swh-metadata-detector']['id']

        metadata = {
            'developmentStatus': None,
            'name': None,
        }
        metadata_rev = {
            'id': self.revision_id_2,
            'translated_metadata': metadata,
            'mappings': ['mapping1'],
            'indexer_configuration_id': tool_id,
        }
        metadata_origin = {
            'origin_id': self.origin_id_1,
            'metadata': metadata,
            'indexer_configuration_id': tool_id,
            'mappings': ['mapping1'],
            'from_revision': self.revision_id_2,
            }

        # when
        self.storage.revision_metadata_add([metadata_rev])

        with self.assertRaises(ValueError):
            self.storage.origin_intrinsic_metadata_add([
                metadata_origin, metadata_origin])

    def test_origin_intrinsic_metadata_search_fulltext(self):
        # given
        tool_id = self.tools['swh-metadata-detector']['id']

        metadata1 = {
            'author': 'John Doe',
        }
        metadata1_rev = {
            'id': self.revision_id_1,
            'translated_metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata1_origin = {
            'origin_id': self.origin_id_1,
            'metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': self.revision_id_1,
        }
        metadata2 = {
            'author': 'Jane Doe',
        }
        metadata2_rev = {
            'id': self.revision_id_2,
            'translated_metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata2_origin = {
            'origin_id': self.origin_id_2,
            'metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': self.revision_id_2,
        }

        # when
        self.storage.revision_metadata_add([metadata1_rev])
        self.storage.origin_intrinsic_metadata_add([metadata1_origin])
        self.storage.revision_metadata_add([metadata2_rev])
        self.storage.origin_intrinsic_metadata_add([metadata2_origin])

        # then
        search = self.storage.origin_intrinsic_metadata_search_fulltext
        self.assertCountEqual(
                [res['origin_id'] for res in search(['Doe'])],
                [self.origin_id_1, self.origin_id_2])
        self.assertEqual(
                [res['origin_id'] for res in search(['John', 'Doe'])],
                [self.origin_id_1])
        self.assertEqual(
                [res['origin_id'] for res in search(['John'])],
                [self.origin_id_1])
        self.assertEqual(
                [res['origin_id'] for res in search(['John', 'Jane'])],
                [])

    def test_origin_intrinsic_metadata_search_fulltext_rank(self):
        # given
        tool_id = self.tools['swh-metadata-detector']['id']

        # The following authors have "Random Person" to add some more content
        # to the JSON data, to work around normalization quirks when there
        # are few words (rank/(1+ln(nb_words)) is very sensitive to nb_words
        # for small values of nb_words).
        metadata1 = {
            'author': [
                'Random Person',
                'John Doe',
                'Jane Doe',
            ]
        }
        metadata1_rev = {
            'id': self.revision_id_1,
            'translated_metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata1_origin = {
            'origin_id': self.origin_id_1,
            'metadata': metadata1,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': self.revision_id_1,
        }
        metadata2 = {
            'author': [
                'Random Person',
                'Jane Doe',
            ]
        }
        metadata2_rev = {
            'id': self.revision_id_2,
            'translated_metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
        }
        metadata2_origin = {
            'origin_id': self.origin_id_2,
            'metadata': metadata2,
            'mappings': [],
            'indexer_configuration_id': tool_id,
            'from_revision': self.revision_id_2,
        }

        # when
        self.storage.revision_metadata_add([metadata1_rev])
        self.storage.origin_intrinsic_metadata_add([metadata1_origin])
        self.storage.revision_metadata_add([metadata2_rev])
        self.storage.origin_intrinsic_metadata_add([metadata2_origin])

        # then
        search = self.storage.origin_intrinsic_metadata_search_fulltext
        self.assertEqual(
                [res['origin_id'] for res in search(['Doe'])],
                [self.origin_id_1, self.origin_id_2])
        self.assertEqual(
                [res['origin_id'] for res in search(['Doe'], limit=1)],
                [self.origin_id_1])
        self.assertEqual(
                [res['origin_id'] for res in search(['John'])],
                [self.origin_id_1])
        self.assertEqual(
                [res['origin_id'] for res in search(['Jane'])],
                [self.origin_id_2, self.origin_id_1])
        self.assertEqual(
                [res['origin_id'] for res in search(['John', 'Jane'])],
                [self.origin_id_1])

    def _fill_origin_intrinsic_metadata(self):
        tool1_id = self.tools['swh-metadata-detector']['id']
        tool2_id = self.tools['swh-metadata-detector2']['id']

        metadata1 = {
            '@context': 'foo',
            'author': 'John Doe',
        }
        metadata1_rev = {
            'id': self.revision_id_1,
            'translated_metadata': metadata1,
            'mappings': ['npm'],
            'indexer_configuration_id': tool1_id,
        }
        metadata1_origin = {
            'origin_id': self.origin_id_1,
            'metadata': metadata1,
            'mappings': ['npm'],
            'indexer_configuration_id': tool1_id,
            'from_revision': self.revision_id_1,
        }
        metadata2 = {
            '@context': 'foo',
            'author': 'Jane Doe',
        }
        metadata2_rev = {
            'id': self.revision_id_2,
            'translated_metadata': metadata2,
            'mappings': ['npm', 'gemspec'],
            'indexer_configuration_id': tool2_id,
        }
        metadata2_origin = {
            'origin_id': self.origin_id_2,
            'metadata': metadata2,
            'mappings': ['npm', 'gemspec'],
            'indexer_configuration_id': tool2_id,
            'from_revision': self.revision_id_2,
        }
        metadata3 = {
            '@context': 'foo',
        }
        metadata3_rev = {
            'id': self.revision_id_3,
            'translated_metadata': metadata3,
            'mappings': ['npm', 'gemspec'],
            'indexer_configuration_id': tool2_id,
        }
        metadata3_origin = {
            'origin_id': self.origin_id_3,
            'metadata': metadata3,
            'mappings': ['pkg-info'],
            'indexer_configuration_id': tool2_id,
            'from_revision': self.revision_id_3,
        }

        self.storage.revision_metadata_add([metadata1_rev])
        self.storage.origin_intrinsic_metadata_add([metadata1_origin])
        self.storage.revision_metadata_add([metadata2_rev])
        self.storage.origin_intrinsic_metadata_add([metadata2_origin])
        self.storage.revision_metadata_add([metadata3_rev])
        self.storage.origin_intrinsic_metadata_add([metadata3_origin])

    def test_origin_intrinsic_metadata_search_by_producer(self):
        self._fill_origin_intrinsic_metadata()
        tool1 = self.tools['swh-metadata-detector']
        tool2 = self.tools['swh-metadata-detector2']
        endpoint = self.storage.origin_intrinsic_metadata_search_by_producer

        # test pagination
        self.assertCountEqual(
            endpoint(ids_only=True),
            [self.origin_id_1, self.origin_id_2, self.origin_id_3])
        self.assertCountEqual(
            endpoint(start=0, ids_only=True),
            [self.origin_id_1, self.origin_id_2, self.origin_id_3])
        self.assertCountEqual(
            endpoint(start=0, limit=2, ids_only=True),
            [self.origin_id_1, self.origin_id_2])
        self.assertCountEqual(
            endpoint(start=self.origin_id_1+1, ids_only=True),
            [self.origin_id_2, self.origin_id_3])
        self.assertCountEqual(
            endpoint(start=self.origin_id_1+1, end=self.origin_id_3-1,
                     ids_only=True),
            [self.origin_id_2])

        # test mappings filtering
        self.assertCountEqual(
            endpoint(mappings=['npm'], ids_only=True),
            [self.origin_id_1, self.origin_id_2])
        self.assertCountEqual(
            endpoint(mappings=['npm', 'gemspec'], ids_only=True),
            [self.origin_id_1, self.origin_id_2])
        self.assertCountEqual(
            endpoint(mappings=['gemspec'], ids_only=True),
            [self.origin_id_2])
        self.assertCountEqual(
            endpoint(mappings=['pkg-info'], ids_only=True),
            [self.origin_id_3])
        self.assertCountEqual(
            endpoint(mappings=['foobar'], ids_only=True),
            [])

        # test pagination + mappings
        self.assertCountEqual(
            endpoint(mappings=['npm'], limit=1, ids_only=True),
            [self.origin_id_1])

        # test tool filtering
        self.assertCountEqual(
            endpoint(tool_ids=[tool1['id']], ids_only=True),
            [self.origin_id_1])
        self.assertCountEqual(
            endpoint(tool_ids=[tool2['id']], ids_only=True),
            [self.origin_id_2, self.origin_id_3])
        self.assertCountEqual(
            endpoint(tool_ids=[tool1['id'], tool2['id']], ids_only=True),
            [self.origin_id_1, self.origin_id_2, self.origin_id_3])

        # test ids_only=False
        self.assertEqual(list(endpoint(mappings=['gemspec'])), [{
            'origin_id': self.origin_id_2,
            'metadata': {
                '@context': 'foo',
                'author': 'Jane Doe',
            },
            'mappings': ['npm', 'gemspec'],
            'tool': tool2,
            'from_revision': self.revision_id_2,
        }])

    def test_origin_intrinsic_metadata_stats(self):
        self._fill_origin_intrinsic_metadata()

        result = self.storage.origin_intrinsic_metadata_stats()
        self.assertEqual(result, {
            'per_mapping': {
                'gemspec': 1,
                'npm': 2,
                'pkg-info': 1,
                'codemeta': 0,
                'maven': 0,
            },
            'total': 3,
            'non_empty': 2,
        })

    def test_indexer_configuration_add(self):
        tool = {
            'tool_name': 'some-unknown-tool',
            'tool_version': 'some-version',
            'tool_configuration': {"debian-package": "some-package"},
        }

        actual_tool = self.storage.indexer_configuration_get(tool)
        self.assertIsNone(actual_tool)  # does not exist

        # add it
        actual_tools = list(self.storage.indexer_configuration_add([tool]))

        self.assertEqual(len(actual_tools), 1)
        actual_tool = actual_tools[0]
        self.assertIsNotNone(actual_tool)  # now it exists
        new_id = actual_tool.pop('id')
        self.assertEqual(actual_tool, tool)

        actual_tools2 = list(self.storage.indexer_configuration_add([tool]))
        actual_tool2 = actual_tools2[0]
        self.assertIsNotNone(actual_tool2)  # now it exists
        new_id2 = actual_tool2.pop('id')

        self.assertEqual(new_id, new_id2)
        self.assertEqual(actual_tool, actual_tool2)

    def test_indexer_configuration_add_multiple(self):
        tool = {
            'tool_name': 'some-unknown-tool',
            'tool_version': 'some-version',
            'tool_configuration': {"debian-package": "some-package"},
        }

        actual_tools = list(self.storage.indexer_configuration_add([tool]))
        self.assertEqual(len(actual_tools), 1)

        new_tools = [tool, {
            'tool_name': 'yet-another-tool',
            'tool_version': 'version',
            'tool_configuration': {},
        }]

        actual_tools = list(self.storage.indexer_configuration_add(new_tools))
        self.assertEqual(len(actual_tools), 2)

        # order not guaranteed, so we iterate over results to check
        for tool in actual_tools:
            _id = tool.pop('id')
            self.assertIsNotNone(_id)
            self.assertIn(tool, new_tools)

    def test_indexer_configuration_get_missing(self):
        tool = {
            'tool_name': 'unknown-tool',
            'tool_version': '3.1.0rc2-31-ga2cbb8c',
            'tool_configuration': {"command_line": "nomossa <filepath>"},
        }

        actual_tool = self.storage.indexer_configuration_get(tool)

        self.assertIsNone(actual_tool)

    def test_indexer_configuration_get(self):
        tool = {
            'tool_name': 'nomos',
            'tool_version': '3.1.0rc2-31-ga2cbb8c',
            'tool_configuration': {"command_line": "nomossa <filepath>"},
        }

        self.storage.indexer_configuration_add([tool])
        actual_tool = self.storage.indexer_configuration_get(tool)

        expected_tool = tool.copy()
        del actual_tool['id']

        self.assertEqual(expected_tool, actual_tool)

    def test_indexer_configuration_metadata_get_missing_context(self):
        tool = {
            'tool_name': 'swh-metadata-translator',
            'tool_version': '0.0.1',
            'tool_configuration': {"context": "unknown-context"},
        }

        actual_tool = self.storage.indexer_configuration_get(tool)

        self.assertIsNone(actual_tool)

    def test_indexer_configuration_metadata_get(self):
        tool = {
            'tool_name': 'swh-metadata-translator',
            'tool_version': '0.0.1',
            'tool_configuration': {"type": "local", "context": "NpmMapping"},
        }

        self.storage.indexer_configuration_add([tool])
        actual_tool = self.storage.indexer_configuration_get(tool)

        expected_tool = tool.copy()
        expected_tool['id'] = actual_tool['id']

        self.assertEqual(expected_tool, actual_tool)

    @pytest.mark.property_based
    def test_generate_content_mimetype_get_range_limit_none(self):
        """mimetype_get_range call with wrong limit input should fail"""
        with self.assertRaises(ValueError) as e:
            self.storage.content_mimetype_get_range(
                start=None, end=None, indexer_configuration_id=None,
                limit=None)

        self.assertEqual(e.exception.args, (
            'Development error: limit should not be None',))

    @pytest.mark.property_based
    @given(gen_content_mimetypes(min_size=1, max_size=4))
    def test_generate_content_mimetype_get_range_no_limit(self, mimetypes):
        """mimetype_get_range returns mimetypes within range provided"""
        self.reset_storage_tables()
        # add mimetypes to storage
        self.storage.content_mimetype_add(mimetypes)

        # All ids from the db
        content_ids = sorted([c['id'] for c in mimetypes])

        start = content_ids[0]
        end = content_ids[-1]

        # retrieve mimetypes
        tool_id = mimetypes[0]['indexer_configuration_id']
        actual_result = self.storage.content_mimetype_get_range(
            start, end, indexer_configuration_id=tool_id)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        self.assertEqual(len(mimetypes), len(actual_ids))
        self.assertIsNone(actual_next)
        self.assertEqual(content_ids, actual_ids)

    @pytest.mark.property_based
    @given(gen_content_mimetypes(min_size=4, max_size=4))
    def test_generate_content_mimetype_get_range_limit(self, mimetypes):
        """mimetype_get_range paginates results if limit exceeded"""
        self.reset_storage_tables()

        # add mimetypes to storage
        self.storage.content_mimetype_add(mimetypes)

        # input the list of sha1s we want from storage
        content_ids = sorted([c['id'] for c in mimetypes])
        start = content_ids[0]
        end = content_ids[-1]

        # retrieve mimetypes limited to 3 results
        limited_results = len(mimetypes) - 1
        tool_id = mimetypes[0]['indexer_configuration_id']
        actual_result = self.storage.content_mimetype_get_range(
            start, end,
            indexer_configuration_id=tool_id, limit=limited_results)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        self.assertEqual(limited_results, len(actual_ids))
        self.assertIsNotNone(actual_next)
        self.assertEqual(actual_next, content_ids[-1])

        expected_mimetypes = content_ids[:-1]
        self.assertEqual(expected_mimetypes, actual_ids)

        # retrieve next part
        actual_results2 = self.storage.content_mimetype_get_range(
            start=end, end=end, indexer_configuration_id=tool_id)
        actual_ids2 = actual_results2['ids']
        actual_next2 = actual_results2['next']

        self.assertIsNone(actual_next2)
        expected_mimetypes2 = [content_ids[-1]]
        self.assertEqual(expected_mimetypes2, actual_ids2)

    @pytest.mark.property_based
    def test_generate_content_fossology_license_get_range_limit_none(self):
        """license_get_range call with wrong limit input should fail"""
        with self.assertRaises(ValueError) as e:
            self.storage.content_fossology_license_get_range(
                start=None, end=None, indexer_configuration_id=None,
                limit=None)

        self.assertEqual(e.exception.args, (
            'Development error: limit should not be None',))

    @pytest.mark.property_based
    def prepare_mimetypes_from(self, fossology_licenses):
        """Fossology license needs some consistent data in db to run.

        """
        mimetypes = []
        for c in fossology_licenses:
            mimetypes.append({
                'id': c['id'],
                'mimetype': 'text/plain',
                'encoding': 'utf-8',
                'indexer_configuration_id': c['indexer_configuration_id'],
            })
        return mimetypes

    @pytest.mark.property_based
    @given(gen_content_fossology_licenses(min_size=1, max_size=4))
    def test_generate_content_fossology_license_get_range_no_limit(
            self, fossology_licenses):
        """license_get_range returns licenses within range provided"""
        self.reset_storage_tables()
        # craft some consistent mimetypes
        mimetypes = self.prepare_mimetypes_from(fossology_licenses)

        self.storage.content_mimetype_add(mimetypes)
        # add fossology_licenses to storage
        self.storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        content_ids = sorted([c['id'] for c in fossology_licenses])

        start = content_ids[0]
        end = content_ids[-1]

        # retrieve fossology_licenses
        tool_id = fossology_licenses[0]['indexer_configuration_id']
        actual_result = self.storage.content_fossology_license_get_range(
            start, end, indexer_configuration_id=tool_id)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        self.assertEqual(len(fossology_licenses), len(actual_ids))
        self.assertIsNone(actual_next)
        self.assertEqual(content_ids, actual_ids)

    @pytest.mark.property_based
    @given(gen_content_fossology_licenses(min_size=1, max_size=4),
           gen_content_mimetypes(min_size=1, max_size=1))
    def test_generate_content_fossology_license_get_range_no_limit_with_filter(
            self, fossology_licenses, mimetypes):
        """This filters non textual, then returns results within range"""
        self.reset_storage_tables()

        # craft some consistent mimetypes
        _mimetypes = self.prepare_mimetypes_from(fossology_licenses)
        # add binary mimetypes which will get filtered out in results
        for m in mimetypes:
            _mimetypes.append({
                'mimetype': 'binary',
                **m,
            })

        self.storage.content_mimetype_add(_mimetypes)
        # add fossology_licenses to storage
        self.storage.content_fossology_license_add(fossology_licenses)

        # All ids from the db
        content_ids = sorted([c['id'] for c in fossology_licenses])

        start = content_ids[0]
        end = content_ids[-1]

        # retrieve fossology_licenses
        tool_id = fossology_licenses[0]['indexer_configuration_id']
        actual_result = self.storage.content_fossology_license_get_range(
            start, end, indexer_configuration_id=tool_id)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        self.assertEqual(len(fossology_licenses), len(actual_ids))
        self.assertIsNone(actual_next)
        self.assertEqual(content_ids, actual_ids)

    @pytest.mark.property_based
    @given(gen_content_fossology_licenses(min_size=4, max_size=4))
    def test_generate_fossology_license_get_range_limit(
            self, fossology_licenses):
        """fossology_license_get_range paginates results if limit exceeded"""
        self.reset_storage_tables()
        # craft some consistent mimetypes
        mimetypes = self.prepare_mimetypes_from(fossology_licenses)

        # add fossology_licenses to storage
        self.storage.content_mimetype_add(mimetypes)
        self.storage.content_fossology_license_add(fossology_licenses)

        # input the list of sha1s we want from storage
        content_ids = sorted([c['id'] for c in fossology_licenses])
        start = content_ids[0]
        end = content_ids[-1]

        # retrieve fossology_licenses limited to 3 results
        limited_results = len(fossology_licenses) - 1
        tool_id = fossology_licenses[0]['indexer_configuration_id']
        actual_result = self.storage.content_fossology_license_get_range(
            start, end,
            indexer_configuration_id=tool_id, limit=limited_results)

        actual_ids = actual_result['ids']
        actual_next = actual_result['next']

        self.assertEqual(limited_results, len(actual_ids))
        self.assertIsNotNone(actual_next)
        self.assertEqual(actual_next, content_ids[-1])

        expected_fossology_licenses = content_ids[:-1]
        self.assertEqual(expected_fossology_licenses, actual_ids)

        # retrieve next part
        actual_results2 = self.storage.content_fossology_license_get_range(
            start=end, end=end, indexer_configuration_id=tool_id)
        actual_ids2 = actual_results2['ids']
        actual_next2 = actual_results2['next']

        self.assertIsNone(actual_next2)
        expected_fossology_licenses2 = [content_ids[-1]]
        self.assertEqual(expected_fossology_licenses2, actual_ids2)


@pytest.mark.db
class IndexerTestStorage(CommonTestStorage, BasePgTestStorage,
                         unittest.TestCase):
    """Running the tests locally.

    For the client api tests (remote storage), see
    `class`:swh.indexer.storage.test_api_client:TestRemoteStorage
    class.

    """
    pass


def test_mapping_names():
    assert set(MAPPING_NAMES) == {m.name for m in MAPPINGS.values()}



from celery import current_app as app

from swh.indexer.metadata import (
    OriginMetadataIndexer, RevisionMetadataIndexer
)
from .test_origin_head import OriginHeadTestIndexer
from .test_metadata import ContentMetadataTestIndexer
from .utils import BASE_TEST_CONFIG


class RevisionMetadataTestIndexer(RevisionMetadataIndexer):
    """Specific indexer whose configuration is enough to satisfy the
       indexing tests.
    """
    ContentMetadataIndexer = ContentMetadataTestIndexer

    def parse_config_file(self, *args, **kwargs):
        return {
            **BASE_TEST_CONFIG,
            'tools': {
                'name': 'swh-metadata-detector',
                'version': '0.0.2',
                'configuration': {
                    'type': 'local',
                    'context': 'NpmMapping'
                }
            }
        }


class OriginMetadataTestIndexer(OriginMetadataIndexer):
    def parse_config_file(self, *args, **kwargs):
        return {
            **BASE_TEST_CONFIG,
            'tools': []
        }

    def _prepare_sub_indexers(self):
        self.origin_head_indexer = OriginHeadTestIndexer()
        self.revision_metadata_indexer = RevisionMetadataTestIndexer()


@app.task
def revision_metadata(*args, **kwargs):
    indexer = RevisionMetadataTestIndexer()
    indexer.run(*args, **kwargs)
    print('REV RESULT=', indexer.results)


@app.task
def origin_intrinsic_metadata(*args, **kwargs):
    indexer = OriginMetadataTestIndexer()
    indexer.run(*args, **kwargs)
